package de.ifgi.unimuenster.noisetube;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;

import de.ifgi.unimuenster.model.BoundingBox;
import de.ifgi.unimuenster.model.Measurement;
import de.ifgi.unimuenster.util.Util;
import de.noiseserver.dto.MeasureSingleDto;
import de.noiseserver.service.FilterRemote;

/**
 * Performance optimizations are made with the help of this class. These methods
 * are thread safe.
 * 
 * @author steffan
 * 
 */
public class NoiseTube {

	private static final BlockingQueue<BoundingBox> boundingBoxQueue = new PriorityBlockingQueue<BoundingBox>(
			50, new BoundingBoxComparator());

	private static final Map<Integer, List<Measurement>> cities = new HashMap<Integer, List<Measurement>>();

	private NoiseTube() {
	};

	/**
	 * Returns a {@link List} a {@link MeasureSingleDto} to a given city id and
	 * filter. If the city id was requested once before the result will be taken
	 * from a hashmap.
	 * 
	 * @param cityID
	 *            the city by id
	 * @param filter
	 *            a {@link FilterRemote} can be <code>null</null>
	 * @return a {@link List} of {@link MeasureSingleDto}
	 * @throws MalformedURLException
	 * @throws IOException
	 */
	public static synchronized List<MeasureSingleDto> getMeasurements(
			int cityID, FilterRemote filter) throws MalformedURLException,
			IOException {

		if (cities.containsKey(cityID)) {

			return Util.convertToDto(cities.get(cityID), filter);
		} else {

			final List<Measurement> requestResult = NoiseTubeParser
					.parseJson(NoiseTubeRequest.getJSON(cityID));

			cities.put(cityID, requestResult);

			return Util.convertToDto(requestResult, filter);

		}

	}

	/**
	 * Returns a {@link List} a {@link MeasureSingleDto} to a given city id and
	 * filter. If the {@link BoundingBox} fits into a {@link BoundingBox} that was requested before the data will be taken
	 * from the existing {@link BoundingBox}.
	 * 
	 * @param boundingBox a {@link BoundingBox} describing the area of interest.
	 * @param filter
	 *            a {@link FilterRemote} can be <code>null</null>
	 * @return a {@link List} of {@link MeasureSingleDto}
	 * @throws MalformedURLException
	 * @throws IOException
	 */
	public static synchronized List<MeasureSingleDto> getMeasurements(
			BoundingBox boundingBox, FilterRemote filter)
			throws MalformedURLException, IOException {

		for (BoundingBox currentBox : boundingBoxQueue) {

			/**
			 * We have the values already in a different(larger) boundingBox
			 */
			if (currentBox.isInside(boundingBox)) {

				boundingBox.increaseUsageCount();

				return Util.convertToDto(
						currentBox.getMeasurements(boundingBox), filter);
			}

		}

		/**
		 * The values are unknown. Lets optimize the queue.
		 */

		if (boundingBoxQueue.size() < 50) {

			final List<Measurement> requestResult = NoiseTubeParser
					.parseJson(NoiseTubeRequest.getJSON(boundingBox));

			boundingBox.setMeasurements(requestResult);

			boundingBoxQueue.offer(boundingBox);

		}
		/**
		 * The queue is full.
		 */
		else {

			boundingBoxQueue.poll();
			boundingBoxQueue.offer(boundingBox);
		}

		return Util.convertToDto(boundingBox.getMeasurements(), filter);

	}

	/**
	 * The comparator for the {@link BoundingBox} queue. It worked with the help of the 
	 * 'usageCount'
	 * 
	 * @author steffan
	 * 
	 */
	private static class BoundingBoxComparator implements
			Comparator<BoundingBox> {

		@Override
		public int compare(final BoundingBox x, final BoundingBox y) {

			if (x.getUsageCount() < y.getUsageCount()) {
				return -1;
			}
			if (x.getUsageCount() > y.getUsageCount()) {
				return 1;
			}

			return 0;
		}

	}

}
