package de.ifgi.unimuenster.util;

import de.ifgi.unimuenster.model.Measurement;
import de.noiseserver.service.FilterRemote;

/**
 * Provides a method to calculate the shortest distance (the great circle) on a sphere.
 * 
 * @author steffan
 * 
 */
public class Orthodrome {

	//flattening constant 
	private static double f = 1.0 / 298.257223563;
	//sphere constant
	private static double a = 6378.137;

	/**
	 * Calculates the shortest distance between the coordinates given by a {@link FilterRemote} and a {@link Measurement}.
	 * As the distances are typically very short - we use the more precise method to ensure the best possible result.
	 * 
	 * @param filter a {@link FilterRemote}
	 * @param measurement a {@link Measurement}
	 * 
	 * @return the distance in kilometers
	 */
	public static double getDistance(FilterRemote filter,
			Measurement measurement) {
		
		if(filter == null)
			throw new IllegalArgumentException("filter can not be null");
		if(measurement == null)
			throw new IllegalArgumentException("measuremnt can not be null");

		double w, x, y, m, n, r2, r3, r1;

		double phiA = filter.getDistLocation().getLongitude();
		double lambdaA = filter.getDistLocation().getLatitude();

		double phiB = measurement.getLongitude().getLongitude();
		double lambdaB = measurement.getLatitude().getLatitude();

		r1 = toRad((phiA - phiB) / 2);
		r2 = toRad((lambdaA + lambdaB) / 2);
		r3 = toRad((lambdaA - lambdaB) / 2);

		x = Math.sin(r3) * Math.sin(r3) * Math.cos(r1) * Math.cos(r1)
				+ Math.cos(r2) * Math.cos(r2) * Math.sin(r1) * Math.sin(r1);
		y = Math.cos(r3) * Math.cos(r3) * Math.cos(r1) * Math.cos(r1)
				+ Math.sin(r2) * Math.sin(r2) * Math.sin(r1) * Math.sin(r1);

		w = Math.atan(Math.sqrt(x / y));

		m = (3 * (Math.sqrt(x * y) / w) - 1) / (2 * y);
		n = (3 * (Math.sqrt(x * y) / w) + 1) / (2 * x);

		return (2 * w * a)
				* (1 + f * m * Math.sin(r2) * Math.sin(r2)
						* (Math.cos(r3) * Math.cos(r3)) - f * n * Math.cos(r2)
						* Math.cos(r2) * Math.sin(r3) * Math.sin(r3));

	}

	/**
	 * Converts degree into radiant.
	 * @param x value in degrees
	 * @return the result in rad
	 */
	private static double toRad(double x) {
		return (Math.PI / 180 * x);
	}

}
