package de.ifgi.unimuenster.util;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import de.ifgi.unimuenster.model.Measurement;
import de.noiseserver.dto.LocationDto;
import de.noiseserver.dto.MeasureSingleDto;
import de.noiseserver.service.FilterRemote;

/**
 * A collection of utility methods. With these methods it is possible to convert a {@link Measurement} into a {@link MeasureSingleDto}.
 * 
 * @author steffan
 * 
 */
public class Util {
	
	/**
	 * Converts a {@link Measurement} into a {@link MeasureSingleDto}. If not lat/lng values available in the original
	 * measurement the result will be <code>null</null>
	 * @param measurement the {@link Measurement} to convert
	 * @return the converted {@link MeasureSingleDto} or <code>null</code>
	 */
	public static MeasureSingleDto convertToDto(Measurement measurement) {
		
		if(measurement == null)
			throw new IllegalArgumentException("measurement can not be null");

		MeasureSingleDto measureDto = new MeasureSingleDto();

		LocationDto locationDto = new LocationDto();
		
		if(measurement.getLongitude() == null || measurement.getLatitude() == null){
			
			return null;
		}
			
		locationDto.setLongitude(measurement.getLongitude().getLongitude());
		
		locationDto.setLatitude(measurement.getLatitude().getLatitude());
		measureDto.setLocationMeasure(locationDto);

		// converting value
		measureDto.setNoise(measurement.getLoudness());

		return measureDto;
	}
	
	/**
	 * Converts a {@link List} of {@link Measurement} to a {@link List} of {@link MeasureSingleDto}. If the measureFilter is not <code>null</code> 
	 * all filter options are taken into account as set. The result can be empty if the measurements are't convertebable or the filter was to
	 * strict.
	 * @param measurements
	 * @param filter can be <code>null</code>
	 * @return a {@link List} of converted {@link MeasureSingleDto}, list can be empty if nothing was to converted or
	 * if the filter was to restrict
	 */
	public static List<MeasureSingleDto> convertToDto(List<Measurement> measurements, FilterRemote filter){
		
		if(measurements == null)
		throw new IllegalArgumentException("measurements can not be null");
		
		List<Measurement> modifiableMeasurements = new ArrayList<Measurement>(measurements);
		
		List<MeasureSingleDto> resultList = new ArrayList<MeasureSingleDto>(measurements.size());
		
		//we have to filter
		if(filter != null){
			
			//check distance buffer
			if(filter.getDistFrom() != null && filter.getDistTo() != null && filter.getDistLocation() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(Orthodrome.getDistance(filter, currentMeasurement) < filter.getDistFrom() || 
							Orthodrome.getDistance(filter, currentMeasurement) > filter.getDistTo())
					{
					iterator.remove();
					}
				
				}
				
			}
			
			//check distance from
			if(filter.getDistFrom() != null && filter.getDistLocation() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(Orthodrome.getDistance(filter, currentMeasurement) < filter.getDistFrom())
					{
					iterator.remove();
					}
				
				}
				
			}
			//check distance to
			if(filter.getDistTo() != null && filter.getDistLocation() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(Orthodrome.getDistance(filter, currentMeasurement) > filter.getDistTo())
					{
					iterator.remove();
					}
				
				}
				
			}
			
			//we have a time frame
			if(filter.getTimeFrom() != null && filter.getTimeTo() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(filter.getTimeFrom().after(noiseTubeDateToJava(currentMeasurement.getMade_at()))
							|| filter.getTimeTo().before(noiseTubeDateToJava(currentMeasurement.getMade_at())))
					{
					iterator.remove();
					}
					
				}
			}
			//just everything after a certain time
			if(filter.getTimeFrom() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(!filter.getTimeFrom().before(noiseTubeDateToJava(currentMeasurement.getMade_at())))
					{
					iterator.remove();
					}
					
				}
			}
			//just everything before a certain time
			if(filter.getTimeTo() != null){
				
				Iterator<Measurement> iterator = modifiableMeasurements.iterator();
				
				while(iterator.hasNext()){
					
					Measurement currentMeasurement = iterator.next();
					
					if(!filter.getTimeTo().after(noiseTubeDateToJava(currentMeasurement.getMade_at())))
					{
					iterator.remove();
					}
					
				}
			}
		
		}
		
		for(Measurement currentMeasurement : modifiableMeasurements){
			
			Object temp = convertToDto(currentMeasurement);
			
			if(temp != null){
			resultList.add((MeasureSingleDto) temp);
			}
		}
		
		return resultList;
	}
	
	/**
	 * Converts the NoiseTube date string (somehow w3c) into a {@link Calendar}.
	 * @param date the NoiseTube data string
	 * @return the converted {@link Calendar}
	 */
	private static Calendar noiseTubeDateToJava(String date){
		
		//YYYY-MM-DDThh:mm:ss Z �hh:mm
		 SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'"); 
	        
		 Date result = null;
	       
	        try {
	            result = formatter.parse(date);
	        }
	        catch(ParseException pe) {}

	        Calendar calendar = Calendar.getInstance();
	        
	        calendar.setTime(result);
	        
	        return calendar;
	}
}
