package de.noisedroid.activities;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.DialogInterface.OnMultiChoiceClickListener;
import android.content.Intent;
import android.graphics.Rect;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.LinearLayout;
import android.widget.ProgressBar;
import android.widget.TextView;
import de.noisedroid.R;
import de.noisedroid.activities.util.SelectRectActivity;
import de.noisedroid.osm.TileInfo;

/**
 * Eine einfache Activity die den manuellen Download von OSM Daten ermglicht. Speichert leider in keinster Weise seinen
 * Zustand...
 * 
 */
public class OSMDownloadActivity extends Activity {
	/**
	 * Hilfsklasse die den Download einer einzelnen Zoomstufe darstellt. Stellt auch eine View bereit, die den
	 * Downloadzustand anzeigt.
	 * 
	 */
	private class Download {
		private int zoom;
		private Rect tileBounds;
		private byte[] downloadBuffer = new byte[2048];
		private View contentView;
		private boolean cancel = false;
		private ProgressBar progress;
		private boolean replace;
		private TextView textInfo;
		private boolean complete = false;

		/**
		 * Konstruktor
		 * 
		 * @param zoom
		 *            Zu ladende Zoomstufe
		 * @param mapBounds
		 *            BoundingBox des herunterzuladenden Bereichs, in E6-Notation
		 * @param replace
		 *            true, falls vorhandenen Daten berschrieben werdne sollen.
		 */
		public Download(final int zoom, Rect mapBounds, boolean replace) {
			this.zoom = zoom;
			this.replace = replace;
			TileInfo topleft = TileInfo.getTileInfoForLonLat(mapBounds.top / 1E6f, mapBounds.left / 1E6f, zoom);
			TileInfo bottomright = TileInfo.getTileInfoForLonLat(mapBounds.bottom / 1E6f, mapBounds.right / 1E6f, zoom);
			tileBounds = new Rect(topleft.x, topleft.y, bottomright.x, bottomright.y);
			tileBounds.sort();
			
			LayoutInflater inflater = (LayoutInflater) OSMDownloadActivity.this
					.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
			contentView = inflater.inflate(R.layout.osmdownload_item, null);

			textInfo = (TextView) contentView.findViewById(R.id.TextViewInfo);
			textInfo.setText(String.format(getString(R.string.zoomstufe_wartet), zoom));

			Button cancelButton = (Button) contentView.findViewById(R.id.ButtonAbbrechen);
			cancelButton.setOnClickListener(new OnClickListener() {
				public void onClick(View v) {
					if (complete) {
						contentView.setVisibility(View.GONE);
					} else {
						cancel = true;
						textInfo.setText(String.format(getString(R.string.zoomstufe_abgebrochen), zoom));
					}
				}
			});

			progress = (ProgressBar) contentView.findViewById(R.id.ProgressBar);
		}

		/**
		 * Gibt die den Download beschreibende View zurck
		 * 
		 * @return
		 */
		public View getView() {
			return contentView;
		}

		/**
		 * Startet den Download aller Kacheln innerhalb der Zoomstufe. Nutzt {@link OSMDownloadActivity#handler} um
		 * whrend des Downloads die View auf dem richtigen Thread zu aktualisieren
		 */
		private void downloadAll() {
			complete = false;
			final AtomicInteger anzSuccess = new AtomicInteger(0);
			final AtomicInteger anzFail = new AtomicInteger(0);
			final int gesamt = (tileBounds.width() + 1) * (tileBounds.height() + 1);
			all: for (int x = tileBounds.left; x <= tileBounds.right; x++) {
				for (int y = tileBounds.top; y <= tileBounds.bottom; y++) {
					if (cancel) {
						break all;
					}
					if (downloadTile(x, y)) {
						anzSuccess.incrementAndGet();
					} else {
						anzFail.incrementAndGet();
					}
					if (!cancel) {
						handler.post(new Runnable() {
							public void run() {
								textInfo.setText(R.string.zoomstufe_wartet + zoom + ", " + anzSuccess.get() + " von " + gesamt
										+ " Kacheln geladen, " + anzFail.get() + " bersprungen");
								progress.setProgress((int) (((anzSuccess.get() + anzFail.get()) / (float) gesamt) * 100f));
							}
						});
					}
				}
			}
			complete = true;
		}

		/**
		 * Ldt einzelne Kachel herunter.
		 * 
		 * @param x
		 * @param y
		 * @return true falls erfolgreich, false falls Verbindungsfehler oder schon vorhanden (je nach
		 *         replace-Eigenschaft)
		 */
		private boolean downloadTile(int x, int y) {
			try {
				String tileString = zoom + "/" + x + "/" + y;
				File file = new File(Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/"
						+ tileString + ".png");
				if (!file.exists()) {
					file.getParentFile().mkdirs();
					file.createNewFile();
				} else {
					if (!replace) {
						return false;
					}
				}

				URL url = new URL("http://tile.openstreetmap.org/" + tileString + ".png");
				URLConnection ucon = url.openConnection();
				InputStream is = ucon.getInputStream();
				int bytesRead = 0;
				FileOutputStream fos = new FileOutputStream(file);

				while ((bytesRead = is.read(downloadBuffer)) > -1) {
					fos.write(downloadBuffer, 0, bytesRead);
				}

				fos.close();
				is.close();
				return true;
			} catch (Exception e) {
				return false;
			}
		}

	}

	private Rect currentRect;
	private Button startButton;
	private TextView zoomText;
	private CheckBox replaceCheck;
	private TextView obenText, untenText, linksText, rechtsText;

	private String[] zoomstufenText = null;
	private boolean[] zoomstufenSelected = null;

	private static final int RECT_REQUEST = 1;
	private Handler handler = new Handler();

	private LinkedBlockingQueue<Download> downloadQueue = new LinkedBlockingQueue<OSMDownloadActivity.Download>();

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.osmdownload);

		zoomstufenText = new String[19];
		zoomstufenSelected = new boolean[19];
		for (int i = 0; i <= 18; i++) {
			zoomstufenText[i] = "Stufe " + i;
		}

		obenText = (TextView) findViewById(R.id.oben);
		untenText = (TextView) findViewById(R.id.unten);
		rechtsText = (TextView) findViewById(R.id.rechts);
		linksText = (TextView) findViewById(R.id.links);

		zoomText = (TextView) findViewById(R.id.zoomstufen);

		replaceCheck = (CheckBox) findViewById(R.id.CheckBoxReplace);

		final LinearLayout downloadList = (LinearLayout) findViewById(R.id.downloadliste);

		startButton = (Button) findViewById(R.id.ButtonStart);
		startButton.setEnabled(false);
		startButton.setOnClickListener(new OnClickListener() {
			public void onClick(View v) {
				if (currentRect != null) {
					for (int i = 0; i < zoomstufenSelected.length; i++) {
						if (zoomstufenSelected[i]) {
							Download download = new Download(i, currentRect, replaceCheck.isChecked());
							downloadQueue.add(download);
							downloadList.addView(download.getView());
						}
					}
				}
			}
		});

		Button chooseRectButton = (Button) findViewById(R.id.ButtonRect);
		chooseRectButton.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				Intent i = new Intent(OSMDownloadActivity.this, SelectRectActivity.class);
				if (currentRect != null) {
					i.putExtra("rectE6", currentRect);
				}
				startActivityForResult(i, RECT_REQUEST);
			}
		});

		Button chooseZoomButton = (Button) findViewById(R.id.ButtonZoom);
		chooseZoomButton.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				showZoomAuswahl();
			}
		});

		updateRectText();
		updateStart();

		// Der Downloadthread.
		new Thread(new Runnable() {
			public void run() {
				try {
					while (true) {
						downloadQueue.take().downloadAll();
					}
				} catch (InterruptedException e) {

				}
			}
		}).start();
	}

	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		super.onActivityResult(requestCode, resultCode, data);
		if (requestCode == RECT_REQUEST && resultCode == RESULT_OK) {
			// Nutzer hat BoundingBox ausgewhlt
			currentRect = data.getParcelableExtra("rectE6");
			updateRectText();
			updateStart();
		}
	}

	private void showZoomAuswahl() {
		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setTitle("Zoomstufen auswhlen");
		builder.setMultiChoiceItems(zoomstufenText, zoomstufenSelected, new OnMultiChoiceClickListener() {

			public void onClick(DialogInterface dialog, int which, boolean isChecked) {
				updateZoomText();
				updateStart();
			}
		});
		builder.setPositiveButton("OK", null);
		builder.setNeutralButton("Standard", new DialogInterface.OnClickListener() {

			public void onClick(DialogInterface dialog, int which) {
				zoomstufenSelected = new boolean[19];
				zoomstufenSelected[0] = true;
				zoomstufenSelected[1] = true;
				zoomstufenSelected[2] = true;
				zoomstufenSelected[12] = true;
				updateZoomText();
				updateStart();
			}
		});
		AlertDialog alert = builder.create();
		alert.show();
	}

	private void updateRectText() {
		if (currentRect != null) {
			obenText.setText(currentRect.top / 1E6f + "");
			untenText.setText(currentRect.bottom / 1E6f + "");
			linksText.setText(currentRect.left / 1E6f + "");
			rechtsText.setText(currentRect.right / 1E6f + "");
		} else {
			obenText.setText("-");
			untenText.setText("-");
			linksText.setText("-");
			rechtsText.setText("-");
		}
	}

	private void updateZoomText() {
		String res = null;
		for (int i = 0; i < zoomstufenSelected.length; i++) {
			if (zoomstufenSelected[i]) {
				if (res == null) {
					res = i + "";
				} else {
					res += ", " + i;
				}
			}
		}
		if (res == null) {
			res = "-";
		}
		zoomText.setText(res);
	}

	private void updateStart() {
		boolean hatZoom = false;
		for (int i = 0; i < zoomstufenSelected.length; i++) {
			if (zoomstufenSelected[i]) {
				hatZoom = true;
				break;
			}
		}

		startButton.setEnabled(currentRect != null && hatZoom);
	}

}
