package de.noisedroid.activities.measurelist.adapter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import android.content.Context;
import de.noisedroid.activities.measurelist.DownloadComparator;
import de.noisedroid.activities.measurelist.ListProvider.OnProgressUpdateListener;
import de.noisedroid.activities.measurelist.gruppierer.LocalGruppierer;
import de.noisedroid.measure.MeasureAbstract;
import de.noisedroid.util.MeasureFilter;

/**
 * Adapter fr die gruppenbasierte Anzeige von Messungen
 * 
 */
public class MeasureExpandableListAdapter extends BaseMeasureExpandableListAdapter<MeasureGroup, LocalGruppierer>
		implements CheckableMeasureAdapter {

	private List<MeasureAbstract> measureCache = new ArrayList<MeasureAbstract>();
	private CheckMode checkMode = CheckMode.NONE;

	public MeasureExpandableListAdapter(Context context, MeasureFilter baseFilter, LocalGruppierer gruppierer,
			DownloadComparator sortierer) {
		super(context, baseFilter, gruppierer, sortierer);
	}

	@Override
	protected void regroupMeasures() {
		List<MeasureGroup> groupsKopie = new ArrayList<MeasureGroup>(groups);
		groups.clear();
		for (BaseMeasureGroup<?> group : groupsKopie) {
			addMeasures(group.getMeasures());
		}
	}

	@Override
	protected void reorderMeasures() {
		for (MeasureGroup group : groups) {
			group.reorderMeasures(sortierer);
		}
		notifyDataSetChanged();
	}

	public void removeMeasure(MeasureAbstract measure) {
		for (MeasureGroup group : groups) {
			group.removeMeasure(measure);
		}
		notifyDataSetChanged();
	}

	public void removeMeasures(List<MeasureAbstract> measureList) {
		for (MeasureAbstract measure : measureList) {
			removeMeasure(measure);
		}
	}

	public void addMeasure(MeasureAbstract measure) {
		internalAddMeasure(measure);
		notifyDataSetChanged();
	}

	private void internalAddMeasure(MeasureAbstract measure) {
		if (baseFilter == null || baseFilter.filter(measure)) {
			int index = gruppierer.getGroupIndex(measure);
			for (MeasureGroup group : groups) {
				if (group.getGroupId() == index) {
					group.addMeasure(measure, sortierer);
					return;
				}
			}
			MeasureGroup group = new MeasureGroup(index, context, this);
			group.setCheckMode(checkMode);
			group.addMeasure(measure, sortierer);
			addGroup(group);
		} else {
			measureCache.add(measure);
		}
	}

	private void addGroup(MeasureGroup group) {
		int index = Collections.binarySearch(groups, group);
		if (index < 0) {
			groups.add(-index - 1, group);
		} else {
			groups.add(index, group);
		}
	}

	public void addMeasures(List<MeasureAbstract> measureList) {
		for (MeasureAbstract measure : measureList) {
			internalAddMeasure(measure);
		}
		notifyDataSetChanged();
	}

	public void setCheckMode(CheckMode mode) {
		checkMode = mode;
		for (MeasureGroup group : groups) {
			group.setCheckMode(mode);
		}
		notifyDataSetChanged();
	}

	public CheckMode getCheckMode() {
		return checkMode;
	}

	public void setChecked(int groupPosition, int childPosition, boolean check) {
		groups.get(groupPosition).setChecked(childPosition, check);
	}

	public void toggleChecked(int groupPosition, int childPosition) {
		boolean check = !groups.get(groupPosition).isChecked(childPosition);
		groups.get(groupPosition).setChecked(childPosition, check);
		notifyDataSetChanged();
	}

	public void setGroupChecked(int groupPosition, boolean check) {
		groups.get(groupPosition).setGroupChecked(check);
	}

	public void toggleGroupChecked(int groupPosition) {
		boolean check = !groups.get(groupPosition).isGroupChecked();
		groups.get(groupPosition).setGroupChecked(check);
		notifyDataSetChanged();
	}

	public List<MeasureAbstract> getCheckedMeasures() {
		List<MeasureAbstract> resultList = new ArrayList<MeasureAbstract>();
		for (MeasureGroup group : groups) {
			resultList.addAll(group.getCheckedMeasures());
		}
		Iterator<MeasureAbstract> iter = resultList.iterator();
		while (iter.hasNext()) {
			if (!MeasureGroup.isCheckable(iter.next(), checkMode)) {
				iter.remove();
			}
		}
		return resultList;
	}

	public void resetChecked() {
		for (MeasureGroup group : groups) {
			group.resetChecked();
		}
		notifyDataSetChanged();
	}

	public void applyFilter(MeasureFilter filter) {
		for (MeasureGroup group : groups) {
			Iterator<MeasureAbstract> iter = group.getMeasures().iterator();
			while (iter.hasNext()) {
				MeasureAbstract measure = iter.next();
				if (!filter.filter(measure)) {
					measureCache.add(measure);
					iter.remove();
				}
			}
		}

		// Liste aller zwischengespeicherten Messungen kopieren, Zwischenspeicher leeren und Kopie wieder versuchen
		// zuzufgen
		// -> Alle dem Filter wieder entsprechenden Messungen kommen in die Liste, andere werden erneut
		// zwischengespeichert.
		List<MeasureAbstract> cacheClone = new ArrayList<MeasureAbstract>(measureCache);
		measureCache.clear();
		addMeasures(cacheClone);

		removeEmptyGroups();
		notifyDataSetChanged();
		resetChecked();
	}

	private void removeEmptyGroups() {
		Iterator<MeasureGroup> iter = groups.iterator();
		while (iter.hasNext()) {
			if (iter.next().getSize() == 0) {
				iter.remove();
			}
		}
	}

	@Override
	public void reload() {

	}

	@Override
	protected void refilterMeasures() {
		applyFilter(baseFilter);
	}

	@Override
	public void progressUpdate() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void setOnProgressUpdateListener(OnProgressUpdateListener listener) {
		// TODO Auto-generated method stub
		
	}

}
