package de.noisedroid.activities.measurelist.map;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;

import android.content.Context;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Rect;
import android.graphics.Shader.TileMode;
import android.graphics.drawable.BitmapDrawable;
import android.widget.Toast;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.MapView;
import com.google.android.maps.Overlay;
import com.google.android.maps.Projection;

import de.noisedroid.R;
import de.noisedroid.activities.measurelist.ListProvider.OnProgressUpdateListener;
import de.noisedroid.community.CommunityMeasureManager;
import de.noisedroid.community.CommunityMeasureManager.Download;
import de.noisedroid.community.CommunityMeasureManager.MeasuresTileReceivedListener;
import de.noisedroid.community.Request.Error;
import de.noisedroid.community.TileInfo;
import de.noisedroid.measure.MeasureAbstract;
import de.noisedroid.util.MeasureFilter;
import de.noisedroid.util.MeasureItemOverlay;
import de.noisedroid.util.MeasureItemOverlay.MeasureOverlayItem;
import de.noiseserver.Commons.DataSourceIdent;

/**
 * Verwaltet die Kartendarstellung von externen Messungsquellen. Stellt dazu Overlays bereit und synchronisiert diese
 * mit einer {@link MapView}
 */
public class CommunityOverlayHolder {
	private class TileLoadedInfo {
		public int x;
		public int y;

		public TileLoadedInfo(int x, int y) {
			this.x = x;
			this.y = y;
		}

		@Override
		public boolean equals(Object o) {
			if (o instanceof TileLoadedInfo) {
				TileLoadedInfo other = (TileLoadedInfo) o;
				return (x == other.x) && (y == other.y);
			} else {
				return super.equals(o);
			}
		}

		@Override
		public int hashCode() {
			return x ^ y;
		}
	}

	private class MeasureLoadingOverlay extends MeasureItemOverlay {
		private TileInfo tileInfo;
		private boolean loading = false;
		private boolean error = false;

		public MeasureLoadingOverlay(Context context, TileLoadedInfo info) {
			super(context);
			tileInfo = new TileInfo(info.x, info.y);
		}

		public void setLoading(boolean loading) {
			this.loading = loading;
		}

		public void setError(boolean error) {
			this.error = error;
		}

		@Override
		public void draw(Canvas canvas, MapView mapView, boolean shadow) {
			if (!shadow) {
				Rect bbox = tileInfo.getPixelBBox(mapView.getProjection());
				if (!Rect.intersects(bbox, canvas.getClipBounds())) {
					return;
				}

				if (loading) {
					bitmapTileLoading.setBounds(bbox);
					bitmapTileLoading.draw(canvas);
				} else if (error) {
					bitmapTileError.setBounds(bbox);
					bitmapTileError.draw(canvas);
				}
				super.draw(canvas, mapView, shadow);
			}
		}

	}

	// Behandelt Messungen die fr die Kartenansicht bestimmt sind. Neue Messungen werden entsprechend des rumlichen
	// Index bereits vorhandenen Overlays zugeordnet und ersetzen deren Daten. Anderfalls werden neue erstellt.
	private MeasuresTileReceivedListener mapListener = new MeasuresTileReceivedListener() {
		public void onReceived(Download download, Collection<MeasureAbstract> measureList, int tileX, int tileY) {
			if (download != null) {
				synchronized (downloadList) {
					downloadList.remove(download);
				}
				progressUpdate();
			}

			if (enabled) {
				TileLoadedInfo info = new TileLoadedInfo(tileX, tileY);
				MeasureLoadingOverlay overlay = overlayMap.get(info);
				if (overlay != null) {
					if (measureList != null) {
						overlay.setMeasures(measureList, basefilter);
					}
					overlay.setLoading(false);
					overlay.setError(measureList == null);
					mapView.postInvalidate();
				}
			}
		}

		public void onCanceled(Download download, int tileX, int tileY, EnumSet<Error> errorSet) {
			if (download != null) {
				synchronized (downloadList) {
					downloadList.remove(download);
				}
				progressUpdate();
			}

			if (enabled) {
				TileLoadedInfo info = new TileLoadedInfo(tileX, tileY);
				MeasureLoadingOverlay overlay = overlayMap.get(info);
				if (overlay != null && overlay.loading) {
					overlay.setLoading(false);
					overlay.setError(true);
					mapView.postInvalidate();
				}

				if (errorSet != null && errorSet.contains(Error.ConnectionError)) {
					mapView.post(new Runnable() {
						public void run() {
							infoToast.setText(R.string.keine_internetverbindung);
							infoToast.show();
						}
					});

				}
			}
		}

		public boolean respondOnMainThread() {
			return false;
		}

	};

	private LinkedHashMap<TileLoadedInfo, MeasureLoadingOverlay> overlayMap = new LinkedHashMap<TileLoadedInfo, MeasureLoadingOverlay>();
	private List<Download> downloadList = new ArrayList<CommunityMeasureManager.Download>();

	private List<Overlay> mapOverlays;

	private CommunityMeasureManager cMeasureManager;
	private MapView mapView;

	private boolean enabled;

	private MeasureFilter basefilter;

	private OnProgressUpdateListener progressUpdateListener;

	private Toast infoToast;

	private static BitmapDrawable bitmapTileLoading;
	private static BitmapDrawable bitmapTileError;

	public CommunityOverlayHolder(DataSourceIdent dataSource, MapView mapView, MeasureFilter basefilter, Toast infoToast) {
		this.infoToast = infoToast;
		this.basefilter = basefilter;
		this.mapView = mapView;
		this.mapOverlays = mapView.getOverlays();
		this.cMeasureManager = CommunityMeasureManager.getInstance(dataSource);
		bitmapTileLoading = new BitmapDrawable(BitmapFactory.decodeResource(mapView.getResources(),
				R.drawable.tile_download));
		bitmapTileError = new BitmapDrawable(
				BitmapFactory.decodeResource(mapView.getResources(), R.drawable.tile_error));
		bitmapTileLoading.setTileModeXY(TileMode.REPEAT, TileMode.REPEAT);
		bitmapTileError.setTileModeXY(TileMode.REPEAT, TileMode.REPEAT);

	}

	public void setOnProgressUpdateListener(OnProgressUpdateListener progressUpdateListener) {
		this.progressUpdateListener = progressUpdateListener;
	}

	private void progressUpdate() {
		if (progressUpdateListener != null) {
			synchronized (downloadList) {
				progressUpdateListener.onProgressUpdate(downloadList.size());
			}
		}
	}

	public int getProgress() {
		return downloadList.size();
	}

	public boolean cancel() {
		List<MeasureLoadingOverlay> waitingList = new ArrayList<MeasureLoadingOverlay>();
		for (MeasureLoadingOverlay measureOverlay : overlayMap.values()) {
			if (measureOverlay.loading) {
				waitingList.add(measureOverlay);
			}
		}
		mapOverlays.removeAll(waitingList);
		overlayMap.values().removeAll(waitingList);
		cancelDownloads();

		return waitingList.size() > 0;
	}

	private void cancelDownloads() {
		synchronized (downloadList) {
			List<Download> downloadListTemp = new ArrayList<CommunityMeasureManager.Download>(downloadList);
			for (Download download : downloadListTemp) {
				download.cancel();
			}
		}
		progressUpdate();
	}

	public void update() {
		if (enabled) {
			updateCommunityOverlays();
		}
	}

	/**
	 * Aktualisiert die MapView Overlay-Liste entsprechend der {@link Projection}, sodass der sichtbare Kartenbereich
	 * mit Overlays abgedeckt wird.
	 * 
	 */
	private void updateCommunityOverlays() {
		Projection proj = mapView.getProjection();
		GeoPoint gPoint = proj.fromPixels(0, 0);
		TileInfo tileInfo = TileInfo.getTileInfoForLonLat(gPoint);
		Rect tileRect = tileInfo.getPixelBBox(proj);
		int anzTiles = 0;
		if (tileRect.width() != 0 && tileRect.height() != 0) {
			int restX = mapView.getWidth() - tileRect.right;
			int spanX = 1 + restX / tileRect.width();
			int restY = mapView.getHeight() - tileRect.bottom;
			int spanY = 1 + restY / tileRect.height();

			int startx = tileInfo.x;
			int starty = tileInfo.y;
			tiles: for (int dy = 0; dy <= spanY; dy++) {
				for (int dx = 0; dx <= spanX; dx++) {
					TileLoadedInfo info = new TileLoadedInfo(startx + dx, starty + dy);
					if (!overlayMap.containsKey(info)) {
						addCommunityOverlayForTile(info);
					}
					anzTiles++;
					if (anzTiles >= 10) {
						break tiles;
					}
				}
			}
		}
	}

	public void setFilter(MeasureFilter filter) {
		this.basefilter = (MeasureFilter) filter.clone();
		cancel();
		reload();
	}

	private void addCommunityOverlayForTile(TileLoadedInfo info) {
		if (overlayMap.size() > 20) {
			Entry<TileLoadedInfo, MeasureLoadingOverlay> entry = overlayMap.entrySet().iterator().next();
			mapOverlays.remove(entry.getValue());
			overlayMap.remove(entry.getKey());
		}
		MeasureLoadingOverlay overlay = new MeasureLoadingOverlay(mapView.getContext(), info);
		overlay.setLoading(true);
		overlayMap.put(info, overlay);
		mapOverlays.add(overlay);
		Download download = cMeasureManager.getMeasuresForTile(info.x, info.y, mapListener, true, basefilter);
		if (download != null) {
			synchronized (downloadList) {
				downloadList.add(download);
			}
			progressUpdate();
		}
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
		if (!enabled) {
			cancelDownloads();
			mapOverlays.removeAll(overlayMap.values());
			overlayMap.clear();
			mapView.postInvalidate();
		} else {
			if (mapView.isShown()) {
				updateCommunityOverlays();
			}
		}
	}

	public boolean isEnabled() {
		return enabled;
	}

	public void reload() {
		if (enabled) {
			mapOverlays.removeAll(overlayMap.values());
			overlayMap.clear();
			updateCommunityOverlays();
			mapView.postInvalidate();
		}
	}

	public List<MeasureAbstract> selectMeasures(Rect selectionRectE6) {
		List<MeasureAbstract> measures = new ArrayList<MeasureAbstract>();

		for (MeasureItemOverlay overlay : overlayMap.values()) {
			for (int i = 0; i < overlay.size(); i++) {
				MeasureOverlayItem item = (MeasureOverlayItem) overlay.getItem(i);
				if (selectionRectE6.contains(item.getPoint().getLongitudeE6(), item.getPoint().getLatitudeE6())) {
					measures.add(item.getMeasure());
				}
			}
		}
		return measures;
	}
}
