package de.noisedroid.community;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;

import android.graphics.Point;
import android.graphics.Rect;
import android.location.Location;
import android.os.Handler;

import com.google.android.maps.GeoPoint;

import de.noisedroid.community.Request.Error;
import de.noisedroid.measure.MeasureAbstract;
import de.noiseserver.Commons.DataSourceIdent;
import de.noiseserver.Commons.OrderAttr;
import de.noiseserver.Commons.OrderDir;

/**
 * Versuch einer Messverwaltung fr Messungen aus der NoiseDroidNoiseMappgingCommunity. Nutzt eine Art rumlichen Index
 * wie fr die OSM-Kartenkacheln, sodass nur ntie Anfragen gettigt werden.
 */
public class CommunityMeasureManager {

	public interface MeasuresReceivedListener {
		boolean respondOnMainThread();

		void onReceived(Download download, Collection<MeasureAbstract> measureList);

		void onCanceled(Download download, EnumSet<Error> errorSet);
	}

	public interface MeasuresTileReceivedListener {
		boolean respondOnMainThread();

		void onReceived(Download download, Collection<MeasureAbstract> measureList, int tileX, int tileY);

		void onCanceled(Download download, int tileX, int tileY, EnumSet<Error> errorSet);

	}

	public interface HasMeasuresReceivedListener {
		boolean respondOnMainThread();

		void onReceived(Download download, boolean hasMeasures);

		void onCanceled(Download download, EnumSet<Error> errorSet);
	}

	/**
	 * Hilfsklasse dessen Objekte die Indizierung zwischengespeicherter Messungen ermglichen
	 * 
	 */
	private class TileDownloadInfo extends TileInfo {
		public long lastDownloadTime;
		public DownloadFilter filter = new DownloadFilter();
		// public HashSet<MeasureAbstract> measures = new HashSet<MeasureAbstract>();
		public List<MeasureAbstract> measures;
		public List<TileDownload> downloads = new ArrayList<CommunityMeasureManager.TileDownload>();

		public TileDownloadInfo(Point tilePos) {
			super(tilePos.x, tilePos.y);
		}

		public void addMeasures(List<MeasureAbstract> measureList, DownloadFilter filter) {
			// TODO verschiedene Filter
			this.filter = filter.clone();
			lastDownloadTime = System.currentTimeMillis();

			measures = measureList;
		}
	}

	public abstract class Download implements Runnable {
		protected boolean cancelable;

		protected boolean canceled;

		public void cancel() {
			downloadThreadPool.remove(this);
			canceled = true;
		}

		public boolean isCanceled() {
			return canceled;
		}

		public boolean isCancelable() {
			return cancelable;
		}

		public void run() {
			if (canceled) {
				return;
			}
		}
	}

	private class TileDownload extends Download {
		private TileDownloadInfo tileDownloadInfo;

		public boolean mainThread;
		public DownloadFilter filter;
		public List<MeasuresTileReceivedListener> listeners = new ArrayList<MeasuresTileReceivedListener>();

		public TileDownload(TileDownloadInfo tileDownloadInfo) {
			this.tileDownloadInfo = tileDownloadInfo;
		}

		@Override
		public void cancel() {
			super.cancel();
			tileDownloadInfo.downloads.remove(this);

			for (MeasuresTileReceivedListener listener : listeners) {
				listener.onCanceled(this, tileDownloadInfo.x, tileDownloadInfo.y, null);
			}
			listeners.clear();
		}

		public void run() {
			super.run();
			tileDownloadInfo.downloads.remove(this);
			ServerRequest request = new ServerRequest();
			final GetMeasuresRequest getMeasuresRequest = new GetMeasuresRequest(filter);
			getMeasuresRequest.setDataSource(dataSource);

			request.addRequest(getMeasuresRequest);
			request.postRequest();
			if (mainThread) {
				mainHandler.post(new Runnable() {
					public void run() {
						processResponse(getMeasuresRequest);
					}
				});
			} else {
				distributionThreadPool.execute(new Runnable() {
					public void run() {
						processResponse(getMeasuresRequest);
					}
				});
			}
		}

		private void processResponse(GetMeasuresRequest getMeasuresRequest) {
			if (getMeasuresRequest.errorSet.isEmpty()) {
				List<MeasureAbstract> receivedMeasures = getMeasuresRequest.getMeasures();
				tileDownloadInfo.addMeasures(receivedMeasures, filter);
				for (MeasuresTileReceivedListener listener : listeners) {
					listener.onReceived(this, receivedMeasures, tileDownloadInfo.x, tileDownloadInfo.y);
				}
			} else {
				for (MeasuresTileReceivedListener listener : listeners) {
					listener.onCanceled(this, tileDownloadInfo.x, tileDownloadInfo.y,getMeasuresRequest.errorSet);
				}
				tileDownloadInfo.addMeasures(null, filter);
			}
			listeners.clear();
		}
	}

	private class UnTiledDownload extends Download {
		public boolean mainThread;
		public DownloadFilter filter;
		public List<MeasuresReceivedListener> listeners = new ArrayList<CommunityMeasureManager.MeasuresReceivedListener>();

		@Override
		public void cancel() {
			super.cancel();

			for (MeasuresReceivedListener listener : listeners) {
				listener.onCanceled(this, null);
			}
			listeners.clear();
		}

		public void run() {
			super.run();
			ServerRequest request = new ServerRequest();
			final GetMeasuresRequest getMeasuresRequest = new GetMeasuresRequest(filter);
			getMeasuresRequest.setDataSource(dataSource);

			request.addRequest(getMeasuresRequest);
			request.postRequest();
			if (mainThread) {
				mainHandler.post(new Runnable() {
					public void run() {
						processResponse(getMeasuresRequest);
					}
				});
			} else {
				distributionThreadPool.execute(new Runnable() {
					public void run() {
						processResponse(getMeasuresRequest);
					}
				});
			}
		}

		private void processResponse(GetMeasuresRequest getMeasuresRequest) {
			if (getMeasuresRequest.errorSet.isEmpty()) {
				List<MeasureAbstract> receivedMeasures = getMeasuresRequest.getMeasures();
				for (MeasuresReceivedListener listener : listeners) {
					listener.onReceived(this, receivedMeasures);
				}
			} else {
				for (MeasuresReceivedListener listener : listeners) {
					listener.onCanceled(this, getMeasuresRequest.errorSet);
				}
			}
			listeners.clear();
		}
	}

	private class HasMeasuresDownload extends Download {
		public boolean mainThread;
		public DownloadFilter filter;
		public List<HasMeasuresReceivedListener> listeners = new ArrayList<HasMeasuresReceivedListener>();

		@Override
		public void cancel() {
			super.cancel();

			for (HasMeasuresReceivedListener listener : listeners) {
				listener.onCanceled(this, null);
			}
			listeners.clear();
		}

		public void run() {
			super.run();
			ServerRequest request = new ServerRequest();
			final HasMeasuresRequest hasMeasuresRequest = new HasMeasuresRequest(filter);
			hasMeasuresRequest.setDataSource(dataSource);

			request.addRequest(hasMeasuresRequest);
			request.postRequest();
			if (mainThread) {
				mainHandler.post(new Runnable() {
					public void run() {
						processResponse(hasMeasuresRequest);
					}
				});
			} else {
				distributionThreadPool.execute(new Runnable() {
					public void run() {
						processResponse(hasMeasuresRequest);
					}
				});
			}
		}

		private void processResponse(HasMeasuresRequest hasMeasuresRequest) {
			if (hasMeasuresRequest.errorSet.isEmpty()) {
				for (HasMeasuresReceivedListener listener : listeners) {
					listener.onReceived(this, hasMeasuresRequest.hasMeasures());
				}
			} else {
				for (HasMeasuresReceivedListener listener : listeners) {
					listener.onCanceled(this, hasMeasuresRequest.errorSet);
				}
			}
			listeners.clear();
		}
	}

	private static HashMap<DataSourceIdent, CommunityMeasureManager> managerMap = new HashMap<DataSourceIdent, CommunityMeasureManager>();
	
	private HashMap<Point, TileDownloadInfo> measureMap = new HashMap<Point, TileDownloadInfo>();
	private ThreadPoolExecutor downloadThreadPool;
	private ThreadPoolExecutor distributionThreadPool;
	private Handler mainHandler = new Handler();
	private DataSourceIdent dataSource;

	public static CommunityMeasureManager getInstance(DataSourceIdent dataSource) {
		 CommunityMeasureManager manager = managerMap.get(dataSource);
		 if(manager!=null) {
			 return manager;
		 } else {
			 manager = new CommunityMeasureManager(dataSource);
			 managerMap.put(dataSource, manager);
			 return manager;
		 }
	}
	
	public static CommunityMeasureManager getInstance() {
		 return getInstance(DataSourceIdent.NOISEDROID);
	}
	
	private CommunityMeasureManager(DataSourceIdent dataSource) {
		this.dataSource = dataSource;
		downloadThreadPool = (ThreadPoolExecutor) Executors.newFixedThreadPool(2);
		distributionThreadPool = (ThreadPoolExecutor) Executors.newFixedThreadPool(2);
	}

	public Download getMeasuresForTile(int x, int y, MeasuresTileReceivedListener callback) {
		return getMeasuresForTile(x, y, callback, true, null);
	}

	public Download getMeasuresForTile(int x, int y, MeasuresTileReceivedListener callback, boolean cancelable) {
		return getMeasuresForTile(x, y, callback, cancelable, new DownloadFilter());
	}

	public Download getMeasuresForTile(int x, int y, final MeasuresTileReceivedListener callback, boolean cancelable,
			DownloadFilter filter) {
		// Filter klonen
		filter = filter.clone();

		// Filter immer rumlich auf diese Kachel beschrnken
		filter.bBoxE6 = new TileInfo(x, y).getLLBBoxE6();

		// Maximale Resultatzahl begrenzen
		if (filter.orderAttr == OrderAttr.NONE) {
			filter.orderAttr = OrderAttr.TIME;
			filter.orderDir = OrderDir.DESC;
		}
		if (filter.limit == null) {
			filter.limit = 500;
		}

		final Point tilePos = new Point(x, y);
		TileDownloadInfo tileDownloadInfo = measureMap.get(tilePos);

		// Prfen ob der Bereich bereits heruntergeladen wurde
		if (tileDownloadInfo != null) {
			// ...Falls ja, altes Ergebnis verwenden - wenn vorhanden
			if (tileDownloadInfo.measures != null) {
				if (callback.respondOnMainThread()) {
					callback.onReceived(null, tileDownloadInfo.measures, tilePos.x, tilePos.y);
				} else {
					final List<MeasureAbstract> measures = tileDownloadInfo.measures;
					distributionThreadPool.execute(new Runnable() {
						public void run() {
							callback.onReceived(null, measures, tilePos.x, tilePos.y);
						}
					});
				}
			}
			if (System.currentTimeMillis() - tileDownloadInfo.lastDownloadTime < 30000
					&& tileDownloadInfo.filter.contains(filter)) {
				// Falls Inhalt relativ neu und vom Filter abgedeckt: keine neue Anfrage ntig -> Abbruch
				// Falls zudem Download bislang nicht erfolgreich scheint: Annahme, dass Download weiterhin nicht
				// erfolgreich sein wird -> Mitteilung das Abbruchs
				if (tileDownloadInfo.measures == null) {
					callback.onCanceled(null, tilePos.x, tilePos.y, null);
				}
				return null;
			}
		} else {
			tileDownloadInfo = new TileDownloadInfo(tilePos);
			measureMap.put(tilePos, tileDownloadInfo);
		}

		boolean newDownload = true;
		for (TileDownload download : tileDownloadInfo.downloads) {
			if (!download.isCanceled() && download.filter.contains(filter)) {
				if (!download.listeners.contains(callback)) {
					download.listeners.add(callback);
				}
				if (!download.cancelable) {
					download.cancelable = cancelable;
				}
				download.mainThread |= callback.respondOnMainThread();
				newDownload = false;
				break;
			}
		}

		if (newDownload) {
			TileDownload downloadInfo = new TileDownload(tileDownloadInfo);
			downloadInfo.listeners.add(callback);
			downloadInfo.cancelable = cancelable;
			downloadInfo.filter = filter;
			downloadInfo.mainThread = callback.respondOnMainThread();
			tileDownloadInfo.downloads.add(downloadInfo);
			addDownload(downloadInfo);
			return downloadInfo;
		} else {
			return null;
		}

	}

	private void addDownload(Download tileDownloadInfo) {
		if (downloadThreadPool.getActiveCount() >= 15) {
			Iterator<Runnable> iter = downloadThreadPool.getQueue().iterator();
			while (iter.hasNext()) {
				Object next = iter.next();
				if (next instanceof Download) {
					Download nextDownload = (Download) next;
					if (nextDownload.cancelable) {
						nextDownload.cancel();
						iter.remove();
						break;
					}
				}
			}
		}
		downloadThreadPool.execute(tileDownloadInfo);
	}

	/**
	 * Aus Geosoft 2 Zeiten...scheint inzwischen unntz?
	 * 
	 * Veranlasst das Herunterladen bzw. die Bereitstellung aller Messungen, die sich in einer festgelegten Umgebung um
	 * einen Punkt befinden. Die Rckgabe der Messungen ist wie bei
	 * {@link #getMeasuresForTile(int, int, MeasuresReceivedListener)} ber ein "Callback" realisiert und kann auch
	 * Messungen auerhalb der exakten Umgebung enthalten.
	 * 
	 * @param latE6
	 * @param lonE6
	 * @param dist
	 *            Distanz in Meter
	 * @param callback
	 * @param cancelable
	 */
	public void getMeasuresForUmgebung(int latE6, int lonE6, float dist, MeasuresTileReceivedListener callback,
			boolean cancelable, DownloadFilter filter) {
		Rect bBoxE6 = new Rect();
		bBoxE6.top = (int) (latE6 + Math.toDegrees(dist / 6137000) * 1E6);
		bBoxE6.bottom = (int) (latE6 - Math.toDegrees(dist / 6137000) * 1E6);
		bBoxE6.right = (int) (lonE6 + Math.toDegrees(dist / 6137000 / Math.cos(Math.toRadians(latE6 / 1E6f))) * 1E6);
		bBoxE6.left = (int) (lonE6 - Math.toDegrees(dist / 6137000 / Math.cos(Math.toRadians(latE6 / 1E6f))) * 1E6);

		GeoPoint gPointTopLeft = new GeoPoint(bBoxE6.top, bBoxE6.left);
		GeoPoint gPointBottomRight = new GeoPoint(bBoxE6.bottom, bBoxE6.right);

		TileInfo tileInfoTopLeft = TileInfo.getTileInfoForLonLat(gPointTopLeft);
		TileInfo tileInfoBottomRight = TileInfo.getTileInfoForLonLat(gPointBottomRight);

		int spanX = 1 + (tileInfoBottomRight.x - tileInfoTopLeft.x);
		int spanY = 1 + (tileInfoBottomRight.y - tileInfoTopLeft.y);

		int startx = tileInfoTopLeft.x;
		int starty = tileInfoTopLeft.y;
		for (int dy = 0; dy <= spanY; dy++) {
			for (int dx = 0; dx <= spanX; dx++) {
				getMeasuresForTile(startx + dx, starty + dy, callback, cancelable, filter);
			}
		}

	}

	/**
	 * @see #getMeasuresForUmgebung(int, int, float, MeasuresReceivedListener, boolean)
	 * @param latE6
	 * @param lonE6
	 * @param dist
	 * @param callback
	 */
	public void getMeasuresForUmgebung(int latE6, int lonE6, float dist, MeasuresTileReceivedListener callback) {
		getMeasuresForUmgebung(latE6, lonE6, dist, callback, true, null);
	}

	public void getMeasuresForUmgebung(Location location, Float dist, MeasuresTileReceivedListener callback,
			boolean cancelable, DownloadFilter filter) {
		getMeasuresForUmgebung((int) (location.getLatitude() * 1E6f), (int) (location.getLongitude() * 1E6f), dist,
				callback, cancelable, filter);
	}

	public Download getMeasures(DownloadFilter filter, MeasuresReceivedListener callback) {
		UnTiledDownload downloadInfo = new UnTiledDownload();
		downloadInfo.listeners.add(callback);
		downloadInfo.cancelable = false;
		downloadInfo.filter = filter;
		downloadInfo.mainThread = callback.respondOnMainThread();
		addDownload(downloadInfo);
		return downloadInfo;
	}

	public Download hasMeasures(final DownloadFilter filter, final HasMeasuresReceivedListener callback) {
		HasMeasuresDownload downloadInfo = new HasMeasuresDownload();
		downloadInfo.listeners.add(callback);
		downloadInfo.cancelable = false;
		downloadInfo.filter = filter;
		downloadInfo.mainThread = callback.respondOnMainThread();
		addDownload(downloadInfo);
		return downloadInfo;
	}

}
