package de.noisedroid.community;

import java.util.EnumSet;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlSerializer;

import de.noiseserver.Commons;

/**
 * Abstrakte Klasse zur Reprsentation einer Teilanfrage, beschreibt Methoden zur Anfrageerstellung und Bearbeitung der
 * Antwort sowie Fehlerbehandlung.
 * 
 * Gibt die konkrete Anfrage Aufschluss darber, dass der Nutzer nicht (mehr) eingeloggt ist, muss in diesem Fall
 * {@link Error#NotLoggedIn} in {@link #errorSet} gesetzt werden.
 * 
 */
public abstract class Request {
	public enum Error {
		TooManyRequests, NotLoggedIn, AlreadyUploaded, MeasureRejected, Unknown, RequestError, ConnectionError, NoResponse
	};

	public abstract String getType();

	protected EnumSet<Error> errorSet = EnumSet.noneOf(Error.class);

	public EnumSet<Error> getErrors() {
		return errorSet;
	}

	public Request() {
		errorSet.add(Error.NoResponse);
	}

	public void asXml(XmlSerializer serializer) throws RequestException {
		try {
			serializer.startTag(null, Commons.NODE_REQUEST);
			serializer.attribute(null, Commons.ATTR_REQUEST_TYPE, getType());
			serializeAttributes(serializer);
			serializer.endTag(null, Commons.NODE_REQUEST);
		} catch (Exception e) {
			throw new RequestException(e.getMessage());
		}
	}

	public abstract void serializeAttributes(XmlSerializer serializer) throws RequestException;

	public void processResponse(XmlPullParser parser) throws RequestException {
		errorSet.remove(Error.NoResponse);
	}

	/**
	 * Interface um das Kommando-Entwurfsmuster in {@link Request#forEachChild(XmlPullParser, TagFunction)} zu nutzen.
	 * 
	 */
	public interface TagFunction {
		public void processTag(XmlPullParser parser) throws Exception;
	}

}
