package de.noisedroid.community;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;

import javax.net.SocketFactory;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.MultihomePlainSocketFactory;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpConnectionParams;
import org.apache.http.params.HttpParams;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserFactory;
import org.xmlpull.v1.XmlSerializer;

import android.os.Build;
import android.util.Xml;
import de.noisedroid.community.Request.Error;
import de.noisedroid.community.Request.TagFunction;
import de.noisedroid.util.Util;
import de.noiseserver.Commons;

/**
 * Stellt eine Anfrage an den Server dar, besteht dabei aus mehreren Teilanfragen, zuzufgen via
 * {@link #addRequest(Request)}.
 * 
 */
public class ServerRequest {
	private HashMap<String, Request> requestMap = new HashMap<String, Request>();
	private boolean requestCompleted = false;

	public void addRequest(Request request) {
		requestMap.put(request.getType(), request);
	}

	private static DefaultHttpClient httpClient;

	private void createHttpClient() {
		SchemeRegistry registry = new SchemeRegistry();
		registry.register(new Scheme("http", PlainSocketFactory.getSocketFactory(), 80));
		registry.register(new Scheme("https", SSLSocketFactory.getSocketFactory(), 443));
		
		HttpParams httpParameters = new BasicHttpParams();
		HttpConnectionParams.setSoTimeout(httpParameters, 10000);
		HttpConnectionParams.setConnectionTimeout(httpParameters, 10000);
		ClientConnectionManager cm = new ThreadSafeClientConnManager(httpParameters, registry);
		httpClient = new DefaultHttpClient(cm, httpParameters);
	}

	/**
	 * Sendet die Anfrage
	 * 
	 * @throws RequestException
	 */
	public void postRequest() {
		try {
			if (requestCompleted) {
				throw new RequestException("Request bereits gesendet");
			}
			if (httpClient == null) {
				createHttpClient();
			}

			HttpPost httpPost = new HttpPost(Commons.SERVER_URL);
			ByteArrayEntity requestEntity = new ByteArrayEntity(createRequestXML().toByteArray());
			requestEntity.setContentType("text/xml");
			httpPost.setEntity(requestEntity);

			HttpResponse response = httpClient.execute(httpPost);
			HttpEntity responseEntity = response.getEntity();
			processResponse(responseEntity.getContent());
			requestCompleted = true;
		} catch (RequestException e) {
			// Fehler beim Verarbeiten der Anfrage oder der Antwort
			for (Request request : requestMap.values()) {
				request.errorSet.add(Error.RequestError);
			}
		} catch (Exception e) {
			// Verbindungsfehler
			for (Request request : requestMap.values()) {
				request.errorSet.add(Error.ConnectionError);
			}
		}
	}

	private ByteArrayOutputStream createRequestXML() throws RequestException {
		try {
			ByteArrayOutputStream stream = new ByteArrayOutputStream();

			XmlSerializer serializer = Xml.newSerializer();
			serializer.setOutput(stream, "UTF-8");
			serializer.setFeature("http://xmlpull.org/v1/doc/features.html#indent-output", Boolean.TRUE);
			serializer.startDocument(null, Boolean.TRUE);

			serializer.startTag(null, Commons.NODE_REQUESTS_WRAPPER);

			// Selbstbeschreibung
			serializer.startTag(null, Commons.NODE_SELFDESCRIPTION);
			serializer.startTag(null, "Time");
			serializer.attribute(null, "value", "" + System.currentTimeMillis());
			serializer.endTag(null, "Time");

			serializer.startTag(null, "Device");
			serializer.attribute(null, "value", "" + Build.PRODUCT);
			serializer.endTag(null, "Device");

			// ...mehr?

			serializer.endTag(null, Commons.NODE_SELFDESCRIPTION);

			for (Request req : requestMap.values()) {
				req.asXml(serializer);
			}
			serializer.endTag(null, Commons.NODE_REQUESTS_WRAPPER);
			serializer.endDocument();
			serializer.flush();
			stream.close();
			return stream;
		} catch (Exception e) {
			throw new RequestException(e.getMessage());
		}
	}

	private void processResponse(InputStream input) throws RequestException {
		try {
			XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
			XmlPullParser parser = factory.newPullParser();
			parser.setInput(input, null);
			parser.next();
			Util.forEachChild(parser, new TagFunction() {
				public void processTag(XmlPullParser parser) throws Exception {

					if (parser.getName().equals(Commons.NODE_RESPONSE)) {
						String type = parser.getAttributeValue(null, Commons.ATTR_REQUEST_TYPE);
						if (type == null) {
							return;
						}
						if (type.equalsIgnoreCase("Error")) {
							// Abwrtskompatibilitt zu fraglichem Serververhalten
							// Sonderfall "allgemeiner Fehler"...Serverantwort gilt hier als unlogisch und ist nicht den
							// ursprnglichen Anfragen zuzuordnen. Deshalb einfache Zuordnung zur ersten Anfrage. Von
							// der Verwendung kombinierter Anfragen wird deshalb abgeraten.

							Iterator<Request> iter = requestMap.values().iterator();
							if (iter.hasNext()) {
								Request request = iter.next();
								request.processResponse(parser);
								if (request.errorSet.contains(Error.NotLoggedIn)) {
									// Festsetzen, dass man wohl nicht eingeloggt ist
									AccountManager.setNotLoggedIn();
								}
							}
							// Util.forEachChild(parser, new TagFunction() {
							// public void processTag(XmlPullParser parser) throws Exception {
							// // Der inkonsistente Weg Fehler mitzuteilen
							// if (parser.getName().equalsIgnoreCase("NotLoggedIn")) {
							// for (Request request : requestMap.values()) {
							// request.errorSet.add(Error.NotLoggedIn);
							// }
							// AccountManager.setNotLoggedIn();
							// } else {
							// for (Request request : requestMap.values()) {
							// request.errorSet.add(Error.Unknown);
							// }
							// }
							// }
							// });

						} else {
							// Der Standardweg,
							// Zuordnung der Antworten zu den Anfragen
							if (requestMap.containsKey(type)) {
								Request request = requestMap.get(type);
								request.processResponse(parser);
								if (request.errorSet.contains(Error.NotLoggedIn)) {
									// Festsetzen, dass man wohl nicht eingeloggt ist
									AccountManager.setNotLoggedIn();
								}
							}
						}

					}

				}
			});
		} catch (Exception e) {
			throw new RequestException(e.getMessage());
		} finally {
			try {
				input.close();
			} catch (IOException e) {

			}
		}
	}
}
