package de.noisedroid.measure;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlSerializer;

import de.noisedroid.measure.MeasureManager.ChangeType;
import de.noisedroid.util.Util;
import de.noiseserver.Commons;

/**
 * Abstrakte Messungsklasse
 * 
 * @author Andr
 * 
 */
public abstract class MeasureAbstract implements Comparable<MeasureAbstract>, XMLSerializable {
	public enum UploadStatus {
		Uploaded, NotUploaded, UploadFailed, UploadRejected, NeMischung
	}

	protected static int nextId;

	protected int id;
	protected int serverId = -1;
	protected Calendar time;
	protected boolean isCommunity = false;
	protected Commons.DataSourceIdent remoteDataSource = null;

	public final void asXML(XmlSerializer serializer) throws MeasureSerializerException {
		asXML(serializer, false);
	}

	public abstract UploadStatus getUploadStatus();

	public int getId() {
		return id;
	}

	public int getServerId() {
		if (isCommunity) {
			return id;
		} else {
			return serverId;
		}
	}
	
	public void setRemoteDataSource(Commons.DataSourceIdent remoteDataSource) {
		this.remoteDataSource = remoteDataSource;
	}
	
	public Commons.DataSourceIdent getRemoteDataSource() {
		return remoteDataSource;
	}

	public void setServerId(int serverId) {
		this.serverId = serverId;
		reportChange(ChangeType.MeasureChanged);
	}

	protected abstract void reportChange(ChangeType type);

	public abstract float getNoise();

	public abstract float getMoral();

	public abstract boolean isValid();

	public abstract void remove();

	public Calendar getTime() {
		return time;
	}

	public void setTime(Calendar time) {
		this.time = time;
	}

	protected static int getNextId() {
		return nextId++;
	}

	public static void setNextId(int id) {
		nextId = id;
	}

	protected void serializeAttributes(XmlSerializer serializer, boolean forServer) throws MeasureSerializerException {
		try {
			serializer.startTag(null, "Time");
			serializer.attribute(null, "value", Util.getTimeToString(time));
			serializer.endTag(null, "Time");
		} catch (Exception e) {
			throw new MeasureSerializerException(e.getMessage());
		}
	}

	protected void setAttributeFromParsingEvent(XmlPullParser parser) throws MeasureParserException {
		try {
			if (parser.getEventType() == XmlPullParser.START_TAG) {
				if (parser.getName().equals("Time")) {
					time = Util.getDateFromString(parser.getAttributeValue(null, "value"));
				}
			}
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}

	}

	/**
	 * Gibt an, ob eine Messung hochgeladen werden kann, also falls noch nicht geschehen und nicht bereits abgewiesen
	 * wurde.
	 * 
	 * @return
	 */
	public boolean isUploadable() {
		return !isCommunity() && getUploadStatus() != UploadStatus.Uploaded
				&& getUploadStatus() != UploadStatus.UploadRejected;
	}

	/**
	 * Gibt an, ob die Messung heruntergeladen wurde oder eine eigene lokale ist.
	 * 
	 * @return
	 */
	public boolean isCommunity() {
		return isCommunity;
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof MeasureAbstract) {
			MeasureAbstract other = (MeasureAbstract) o;
			return other.isCommunity == this.isCommunity && other.id == this.id;
		} else
			return false;
	}

	@Override
	public int hashCode() {
		return !isCommunity ? id : -id;
	}

	public abstract void setUploadStatus(UploadStatus uploaded);

}
