package de.noisedroid.measure;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.List;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserFactory;
import org.xmlpull.v1.XmlSerializer;

import android.app.Activity;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.Environment;
import android.os.IBinder;
import android.util.Xml;
import android.widget.Toast;

public class MeasureManager implements MeasureParentSingle, MeasureParentSeries {
	public enum ChangeType {
		MeasureAdded, MeasureRemoved, MeasureChanged, MeasureAddedSeries, MeasureRemovedSeries, MeasureChangedSeries
	};

	public interface MessungsServiceReady {
		void onReady(MessungsService mMessungsService);
	}

	public interface MeasuresChangedListener {
		void onChange(ChangeType type, MeasureAbstract measure);
	}

	private static Activity contextActivity;
	private static MeasureManager instanz;

	private boolean nderungen = false;
	private List<MeasureAbstract> measureStore = new ArrayList<MeasureAbstract>();
	private MessungsServiceReady callback;
	private MessungsService mMessungsService;
	private List<MeasuresChangedListener> changedListeners = new ArrayList<MeasuresChangedListener>(4);

	// Attribute um die Messserien zu managen
	// private boolean isSeriesActive;

	private ServiceConnection mConnection = new ServiceConnection() {
		public void onServiceConnected(ComponentName name, IBinder service) {
			mMessungsService = ((MessungsService.MessungsServiceBinder) service).getService();

			callback.onReady(mMessungsService);
		}

		public void onServiceDisconnected(ComponentName name) {
			mMessungsService = null;
		}
	};

	private MeasureManager() {
		try {
			loadMeasures();
			nderungen = false;
		} catch (MeasureParserException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static void initMessungsVerwaltung(Activity context) {
		MeasureManager.contextActivity = context;
		ComponentName comp = contextActivity.startService(new Intent(contextActivity, MessungsService.class));
		if (comp == null) {
			// TODO Fehlerbehandlung
		}
	}

	public static MeasureManager getInstance() {
		if (instanz == null) {
			instanz = new MeasureManager();
		}
		return instanz;
	}

	public static void stopMessungsVerwaltung() {
		if (instanz != null) {
			instanz.messungsServiceTrennen();
			if (instanz.nderungen) {

				instanz.saveMeasureStore();
			}
			// instanz = null;
		}
	}

	private void messungsServiceTrennen() {
		if (mMessungsService != null) {
			contextActivity.unbindService(mConnection);
			mMessungsService = null;
		}
	}

	private void messungsServiceVerbinden() {
		if (mMessungsService == null) {
			contextActivity.bindService(new Intent(contextActivity, MessungsService.class), mConnection,
					Context.BIND_AUTO_CREATE);
		}
	}

	public void getMessungsService(MessungsServiceReady callback) {
		if (mMessungsService == null) {
			this.callback = callback;
			messungsServiceVerbinden();
		} else {
			callback.onReady(mMessungsService);
		}
	}

	public void addMeasure(MeasureSeries series) {
		if (!series.isValid())
			return;
		int index = Collections.binarySearch(measureStore, series);
		if (index < 0) {
			measureStore.add(-index - 1, series);
		} else {
			measureStore.add(index, series);
		}
		series.setMeasureParent(this);
		reportChange(ChangeType.MeasureAdded, series);
	}

	public void addMeasure(MeasureAutomatic measure, boolean ignoreSeries) {
		if (!measure.isValid())
			return;

		if (MeasureSeries.getActiveSeries() != null) {
			MeasureSeries.getActiveSeries().checkEndLocation(measure.location);
			MeasureSeries.getActiveSeries().checkEndTime();
		}

		if (MeasureSeries.getActiveSeries() != null && !ignoreSeries) {
			// Falls eine Serie aktiv ist, werden alle Messungen der aktiven Serie hinzugefgt
			MeasureSeries.getActiveSeries().addMeasure(measure);
		} else {
			int index = Collections.binarySearch(measureStore, measure);
			if (index < 0) {
				measureStore.add(-index - 1, measure);
			} else {
				measureStore.add(index, measure);
			}
			measure.setMeasureParent(this);
			reportChange(ChangeType.MeasureAdded, measure);
		}

	}

	public void addMeasure(MeasureAutomatic measure) {
		addMeasure(measure, false);
	}

	public void removeMeasure(MeasureAutomatic measure) {
		measureStore.remove(measure);
		reportChange(ChangeType.MeasureRemoved, measure);
	}

	public void removeMeasure(MeasureSeries measure) {
		measureStore.remove(measure);
		reportChange(ChangeType.MeasureRemoved, measure);
	}

	public void addChangeListener(MeasuresChangedListener listener) {
		if (!changedListeners.contains(listener)) {
			changedListeners.add(listener);
		}
	}

	public void removeChangeListener(MeasuresChangedListener listener) {
		changedListeners.remove(listener);
	}

	public void reportChange(ChangeType type, MeasureAbstract measure) {
		nderungen = true;
		for (MeasuresChangedListener listener : changedListeners) {
			if (listener != null) {
				listener.onChange(type, measure);
			}
		}
	}

	public void reportChange(ChangeType type, MeasureSeries measure) {
		reportChange(type, (MeasureAbstract) measure);
	}

	public void reportChange(ChangeType type, MeasureAutomatic measure) {
		reportChange(type, (MeasureAbstract) measure);
	}

	public List<MeasureAbstract> getMeasures() {
		return measureStore;
	}

	private ByteArrayOutputStream createXMLOutputStream() throws IOException, MeasureSerializerException {
		ByteArrayOutputStream stream = new ByteArrayOutputStream();
		XmlSerializer serializer = Xml.newSerializer();

		serializer.setOutput(stream, "UTF-8");
		serializer.setFeature("http://xmlpull.org/v1/doc/features.html#indent-output", Boolean.TRUE);
		serializer.startDocument(null, Boolean.TRUE);
		serializer.startTag(null, "Measures");
		for (MeasureAbstract messung : measureStore) {
			messung.asXML(serializer);
		}
		serializer.endTag(null, "Measures");
		serializer.endDocument();
		serializer.flush();
		stream.close();

		return stream;
	}

	public void saveMeasureStore() {
		File newxmlfile = new File(Environment.getExternalStorageDirectory()
				+ "/Android/data/de.noisedroid/measures.xml");
		try {
			if (!newxmlfile.exists()) {
				newxmlfile.getParentFile().mkdirs();
			}
			newxmlfile.createNewFile();
			byte[] byteArray = createXMLOutputStream().toByteArray();
			FileOutputStream fos = new FileOutputStream(newxmlfile);
			fos.write(byteArray);
			fos.close();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			Toast.makeText(contextActivity, "Fehler beim Speichern", Toast.LENGTH_LONG).show();
		}
		// Toast.makeText(contextActivity, "erfolgreich gespeichert", Toast.LENGTH_LONG).show();
	}

	/**
	 * Parser fr Messungen.
	 * 
	 * @throws MeasureParserException
	 */
	private void loadMeasures() throws MeasureParserException {
		try {
			XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
			XmlPullParser parser = factory.newPullParser();
			Reader reader = new FileReader(Environment.getExternalStorageDirectory()
					+ "/Android/data/de.noisedroid/measures.xml");
			parser.setInput(reader);
			int eventType = parser.getEventType();
			// eventType = parser.next();
			while (eventType != XmlPullParser.END_DOCUMENT) {
				if (parser.getName() != null && parser.getEventType() == XmlPullParser.START_TAG) {
					if (parser.getName().equals("Measure")
							&& parser.getAttributeValue(null, "type").equalsIgnoreCase("automatic")) {
						addMeasure(MeasureAutomatic.fromXML(parser));
					} else if (parser.getName().equals("Measure")
							&& parser.getAttributeValue(null, "type").equalsIgnoreCase("manual")) {
						addMeasure(MeasureManual.fromXML(parser));
					} else if (parser.getName().equals("MeasureSeries")) {
						addMeasure(MeasureSeries.fromXML(parser));
					} else {
						eventType = parser.next();
						continue;
					}
				}
				eventType = parser.next();
			}
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}

	}

	public MeasureAbstract findById(int id) {
		// MeasureSeries sind die ersten Eintrge und werden entsprechend durchsucht
		for (MeasureAbstract measure : measureStore) {
			if (measure instanceof MeasureSeries) {
				// Prfen der Serie an sich
				if (measure.getId() == id) {
					return measure;
				}

				// Prfen der Serieninhalte
				MeasureAutomatic m = ((MeasureSeries) measure).findById(id);
				if (m != null) {
					return m;
				}
			} else {
				// Alle Serien abgesucht
				break;
			}
		}

		// Durchsuchen des Rests ber binre Suche
		int index = Collections.binarySearch(measureStore, new MeasureSearchDummy(id));
		if (index >= 0) {
			return measureStore.get(index);
		} else {
			return null;
		}
	}

	public MeasureSeries findSeriesById(int id) {
		// MeasureSeries sind die ersten Eintrge und werden entsprechend durchsucht
		for (MeasureAbstract measure : measureStore) {
			if (measure instanceof MeasureSeries) {
				if (measure.getId() == id) {
					return (MeasureSeries) measure;
				}
			} else {
				return null;
			}

		}
		return null;
	}

	public MeasureSeries getLatestSeries() {
		if (measureStore.size() > 0 && measureStore.get(0) instanceof MeasureSeries) {
			return (MeasureSeries) measureStore.get(0);
		} else {
			return null;
		}
	}

	public List<MeasureSeries> getSeries() {
		List<MeasureSeries> seriesList = new ArrayList<MeasureSeries>();
		for (MeasureAbstract measure : measureStore) {
			if (measure instanceof MeasureSeries) {
				seriesList.add((MeasureSeries) measure);
			} else {
				break;
			}
		}
		return seriesList;
	}

}
