package de.noisedroid.measure;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlSerializer;

import android.location.Location;
import de.noisedroid.measure.MeasureManager.ChangeType;
import de.noiseserver.Commons.DataSourceIdent;

/**
 * Manuelle Messung. Abgeleitet von der automatischen Messung, enthlt aber noch zustzliche Attribute.
 * 
 * @author Andr
 * 
 */
public class MeasureManual extends MeasureAutomatic {

	protected float rating;
	protected float distToNoise;
	protected Location locationSource;
	protected String description = "";
	protected List<String> tags = new ArrayList<String>();

	public MeasureManual() {
		super();

	}

	public List<String> getTags() {
		return tags;
	}

	public void setTags(List<String> tags) {
		this.tags = tags;
	}

	public void addTag(String tag) {
		tags.add(tag);
	}

	public MeasureManual(Location location, Calendar time, float noise, float moral) {
		super(location, time, noise, moral);
	}

	private MeasureManual(int id) {
		this.id = id;
	}

	public void setRating(float rating) {
		this.rating = rating;
		reportChange(ChangeType.MeasureChanged);
	}

	public void setDistToNoise(float dist) {
		distToNoise = dist;
		reportChange(ChangeType.MeasureChanged);
	}

	public float getDistToNoise() {
		return distToNoise;
	}

	public float getRating() {
		return rating;

	}

	public void setLocationSource(Location locationSource) {
		this.locationSource = locationSource;
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
		reportChange(ChangeType.MeasureChanged);
	}

	@Override
	protected void serializeAttributes(XmlSerializer serializer, boolean forServer) throws MeasureSerializerException {
		try {
			super.serializeAttributes(serializer, forServer);
			serializer.startTag(null, "Description");
			serializer.text("" + description);
			serializer.endTag(null, "Description");
			serializer.startTag(null, "Distance");
			serializer.attribute(null, "value", "" + distToNoise);
			serializer.endTag(null, "Distance");
			serializer.startTag(null, "Rating");
			serializer.attribute(null, "value", "" + rating);
			serializer.endTag(null, "Rating");
			serializer.startTag(null, "Tags");
			for (String element : tags) {
				serializer.startTag(null, "Tag");
				serializer.attribute(null, "value", "" + element);
				serializer.endTag(null, "Tag");
			}
			serializer.endTag(null, "Tags");

			if (locationSource != null) {
				serializer.startTag(null, "LocationSource");
				serializer.attribute(null, "longitude", "" + location.getLongitude());
				serializer.attribute(null, "latitude", "" + location.getLatitude());
				serializer.endTag(null, "LocationSource");
			}

		} catch (Exception e) {
			throw new MeasureSerializerException(e.getMessage());
		}
	}

	public void asXML(XmlSerializer serializer, boolean forServer) throws MeasureSerializerException {
		try {
			serializer.startTag(null, "Measure");
			serializer.attribute(null, "id", "" + id);
			serializer.attribute(null, "type", "manual");

			serializeAttributes(serializer, forServer);

			serializer.endTag(null, "Measure");
		} catch (Exception e) {
			throw new MeasureSerializerException(e.getMessage());
		}
	}

	@Override
	protected void setAttributeFromParsingEvent(XmlPullParser parser) throws MeasureParserException {
		try {
			if (parser.getEventType() == XmlPullParser.START_TAG) {
				if (parser.getName().equals("Description")) {
					description = parser.nextText();
				} else if (parser.getName().equals("Distance")) {
					distToNoise = Float.parseFloat(parser.getAttributeValue(null, "value"));
				} else if (parser.getName().equals("Rating")) {
					rating = Float.parseFloat(parser.getAttributeValue(null, "value"));
				} else if (parser.getName().equals("Tag")) {
					tags.add(parser.getAttributeValue(null, "value"));
				} else if (parser.getName().equals("LocationSource")) {
					locationSource = new Location("user");
					locationSource.setLongitude(Double.parseDouble(parser.getAttributeValue(null, "longitude")));
					locationSource.setLatitude(Double.parseDouble(parser.getAttributeValue(null, "latitude")));
				} else {
					super.setAttributeFromParsingEvent(parser);
				}
			}
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}

	}

	public static MeasureManual fromXML(XmlPullParser parser) throws MeasureParserException {
		return fromXML(parser, null);
	}

	public static MeasureManual fromXML(XmlPullParser parser, DataSourceIdent dataSource) throws MeasureParserException {
		try {
			int id;
			if (dataSource == DataSourceIdent.NOISETUBE) {
				id = 0;
			} else {
				id = Integer.parseInt(parser.getAttributeValue(null, "id"));
			}
			MeasureManual m = new MeasureManual(id);
			m.isCommunity = dataSource != null;
			m.setRemoteDataSource(dataSource);

			while (true) {
				if (parser.getName() != null) {
					try {
						m.setAttributeFromParsingEvent(parser);
					} catch (MeasureParserException e) {
						// Fehlerhaftes Attribut
					}
					if (parser.getEventType() == XmlPullParser.END_TAG && parser.getName().equals("Measure")) {
						break;
					}
				}
				parser.next();
			}
			if (!m.isCommunity && m.id >= nextId) {
				setNextId(m.id + 1);
			}
			return m;
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}

	}

	@Override
	public String toString() {
		return super.toString() + "\n" + "Description: " + description + "\n" + "Distance: " + distToNoise + "\n"
				+ "Rating: " + rating;

	}

	/**
	 * Berechnet den Schalldruck in Abhngigkeit der Entfernung. Der effektive Noise Wert entspricht hier dem
	 * Schalldruck in 1m Entfernung. Quelle: http://www.sengpielaudio.com/Rechner-entfernung.htm
	 * 
	 * @return effektiver Schalldruck in dB in 1m Entfernung
	 */
	public float getEffectiveNoise() {
		if (distToNoise > 0) {
			return (float) (noise + 20 * Math.log10(distToNoise));
		} else if (distToNoise == 0 || distToNoise == -1) {
			return noise;
		}

		else {
			throw new IllegalStateException("Distance required");
		}
	}

}
