package de.noisedroid.measure;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.List;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlSerializer;

import android.graphics.Point;
import android.graphics.Rect;
import android.location.Location;
import de.noisedroid.community.UpdateSeriesRequest;
import de.noisedroid.measure.MeasureManager.ChangeType;
import de.noisedroid.measure.MeasureManager.MeasuresChangedListener;
import de.noisedroid.util.Noise;
import de.noisedroid.util.Util;
import de.noiseserver.Commons.DataSourceIdent;

/**
 * Messserie, erweitert die abstrakte Messungsklasse. Eine Messserie besteht aus automatischen Messungsobjekten.
 * 
 * @author Andr
 * 
 */
public class MeasureSeries extends MeasureAbstract implements MeasureParentSingle {

	private static MeasureSeries activeSeries;

	protected Calendar endTime = null;
	protected Location endLocation = null;
	public static final float endLocationRadius = 200;
	protected String description = "";
	protected String title = "";
	protected List<MeasureAutomatic> measures = new ArrayList<MeasureAutomatic>();
	protected MeasureParentSeries measureParent;
	protected List<String> tags = new ArrayList<String>();

	private List<MeasuresChangedListener> changedListeners = new ArrayList<MeasuresChangedListener>(4);

	public MeasureSeries(String description) {
		this();
		this.description = description;
	}

	public MeasureSeries() {
		this(getNextId());
		Calendar t = new GregorianCalendar();
		time = t;
	}

	public void setTags(List<String> tags) {
		this.tags = tags;
		reportChange(ChangeType.MeasureChanged);
	}

	public List<String> getTags() {
		return tags;
	}

	private MeasureSeries(int id) {
		this.id = id;
	}

	public boolean isActive() {
		return activeSeries != null && activeSeries == this;
	}

	public static MeasureSeries getActiveSeries() {
		if (activeSeries != null) {
			activeSeries.checkEndTime();
		}

		return activeSeries;
	}

	public void setActive(boolean active) {
		if (active) {
			if (activeSeries != null && activeSeries != this) {
				MeasureSeries tempActive = activeSeries;
				activeSeries = this;
				tempActive.reportChange(ChangeType.MeasureChanged);
			} else {
				activeSeries = this;
			}
		} else {
			if (activeSeries == this) {
				activeSeries = null;
			}
		}
		checkEndTime();
		this.reportChange(ChangeType.MeasureChanged);
	}

	public void checkEndTime() {
		// Zeitliche Beschrnkung der aktiven Serie prfen
		if (!isActive()) {
			return;
		}
		if (endTime != null) {
			if (new GregorianCalendar().after(endTime)) {
				setActive(false);
			}
		}
	}

	public void checkEndLocation(Location location) {
		// Rumliche Beschrnkung der aktiven Serie prfen
		if (!isActive()) {
			return;
		}
		if (endLocation != null) {
			if (location.distanceTo(endLocation) <= MeasureSeries.endLocationRadius) {
				setActive(false);
			}
		}
	}

	@Override
	public boolean isValid() {
		if (time == null)
			return false;

		for (MeasureAutomatic measure : measures) {
			if (!measure.isValid()) {
				return false;
			}
		}
		return true;
	}

	public void addMeasure(MeasureAutomatic measure) {
		if (measure.isValid()) {
			int index = Collections.binarySearch(measures, measure);
			if (index < 0) {
				measures.add(-index - 1, measure);
			} else {
				measures.add(index, measure);
			}
			measure.setMeasureParent(this);
			reportChange(ChangeType.MeasureAdded, measure);
		}
	}

	public void removeMeasure(MeasureAutomatic measure) {
		measures.remove(measure);
		reportChange(ChangeType.MeasureRemoved, measure);
	}

	public void remove() {
		if (activeSeries == this) {
			activeSeries = null;
		}
		measureParent.removeMeasure(this);
	}

	public List<MeasureAutomatic> getMeasures() {
		return measures;
	}

	public String getDescription() {
		return description;
	}

	public Calendar getEndTime() {
		return endTime;
	}

	public void setEndtime(Calendar end) {
		endTime = end;
		reportChange(ChangeType.MeasureChanged);
	}

	public void setDescription(String description) {
		this.description = description;
		reportChange(ChangeType.MeasureChanged);
	}

	public void setTitle(String title) {
		this.title = title;
		reportChange(ChangeType.MeasureChanged);
	}

	public String getTitle() {
		return title;
	}

	public boolean hasMeasureParent() {
		return measureParent != null;
	}

	public void setMeasureParent(MeasureParentSeries measureParent) {
		this.measureParent = measureParent;
	}

	public void reportChange(ChangeType type, MeasureAutomatic measure) {
		if (measureParent != null) {
			if (type == ChangeType.MeasureAdded)
				measureParent.reportChange(ChangeType.MeasureAddedSeries, measure);
			if (type == ChangeType.MeasureRemoved)
				measureParent.reportChange(ChangeType.MeasureRemovedSeries, measure);
			if (type == ChangeType.MeasureChanged)
				measureParent.reportChange(ChangeType.MeasureChangedSeries, measure);
		}
		for (MeasuresChangedListener listener : changedListeners) {
			if (listener != null) {
				listener.onChange(type, measure);
			}
		}
	}

	protected void reportChange(ChangeType type) {
		if (measureParent != null) {
			measureParent.reportChange(type, this);
		}
		for (MeasuresChangedListener listener : changedListeners) {
			if (listener != null) {
				listener.onChange(type, this);
			}
		}
	}

	@Override
	public float getNoise() {
		float sumNoise = 0;
		for (MeasureAutomatic measure : measures) {
			sumNoise += Noise.DBToPcm(measure.getNoise());
		}
		return Noise.pcmToDB(sumNoise / measures.size());
	}

	public float getMaxNoise() {
		float maxNoise = 0;
		for (MeasureAutomatic measure : measures) {
			if (measure.getNoise() > maxNoise) {
				maxNoise = measure.getNoise();
			}
		}
		return maxNoise;
	}

	public float getMinNoise() {
		if (measures.size() == 0) {
			return 0;
		} else {
			float minNoise = Float.MAX_VALUE;
			for (MeasureAutomatic measure : measures) {
				if (measure.getNoise() < minNoise) {
					minNoise = measure.getNoise();
				}
			}
			return minNoise;
		}

	}

	public float getMoral() {
		float sumMoral = 0;
		for (MeasureAutomatic measure : measures) {
			sumMoral += measure.getMoral();
		}
		return (sumMoral / measures.size());
	}

	public double getMaxMoral() {
		double maxMoral = 0;
		for (MeasureAutomatic measure : measures) {
			if (measure.getMoral() > maxMoral) {
				maxMoral = measure.getMoral();
			}
		}
		return maxMoral;
	}

	public double getMinMoral() {
		if (measures.size() == 0) {
			return 0;
		} else {
			double minMoral = Double.MAX_VALUE;
			for (MeasureAutomatic measure : measures) {
				if (measure.getMoral() < minMoral) {
					minMoral = measure.getMoral();
				}
			}
			return minMoral;
		}
	}

	public Location getEndLocation() {
		return endLocation;
	}

	public void setEndLocation(Location end) {
		endLocation = end;
	}

	public Point getCenterLocationE6() {
		Rect bounds = null;
		for (MeasureAutomatic measure : measures) {
			int lonE6 = (int) (measure.getLocation().getLongitude() * 1E6);
			int latE6 = (int) (measure.getLocation().getLatitude() * 1E6);
			if (bounds == null) {
				bounds = new Rect(lonE6, latE6, lonE6, latE6);
			} else {
				bounds.union(lonE6, latE6);
			}
		}
		if (bounds == null) {
			return null;
		} else {
			return new Point(bounds.centerX(), bounds.centerY());
		}
	}

	@Override
	public UploadStatus getUploadStatus() {
		UploadStatus status = null;
		for (MeasureAutomatic measure : measures) {
			if (status == null) {
				status = measure.getUploadStatus();
			} else if (status != measure.getUploadStatus()) {
				return UploadStatus.NeMischung;
			}
		}
		if (status == null) {
			return UploadStatus.NotUploaded;
		} else {
			return status;
		}
	}

	@Override
	public void setUploadStatus(UploadStatus uploaded) {
		if (uploaded == UploadStatus.UploadFailed || uploaded == UploadStatus.UploadRejected) {
			for (MeasureAutomatic measure : measures) {
				if (measure.getUploadStatus() == UploadStatus.NotUploaded) {
					measure.setUploadStatus(uploaded);
				} else if (measure.getUploadStatus() == UploadStatus.UploadFailed
						&& uploaded == UploadStatus.UploadRejected) {
					measure.setUploadStatus(uploaded);
				}
			}
		}
	}

	public boolean isSeriesContainerUploaded() {
		return serverId > 0;
	}

	public void asXML(XmlSerializer serializer, boolean forServer) throws MeasureSerializerException {
		asXML(serializer, measures, forServer, false);
	}

	/**
	 * Spezielle Version des {@link #asXML(XmlSerializer, boolean)} die es erlaubt die Serie mit einer speziellen
	 * Auswahl an Teilmessungen zu serialisieren. Wichtig fr das "partielle Hochladen".
	 * 
	 * @param serializer
	 * @param measureList
	 * @param forServer
	 *            Gibt an, dass die Serialisierung zum Versenden an den Server geadacht ist. Schliet fr die interne
	 *            Verwaltung ntige Attribute aus.
	 * @param forUpdate
	 *            Gibt an, dass die Serialisierung fr die serverseitige Serienaktualisierung gedacaht ist (
	 *            {@link UpdateSeriesRequest}). Schliet alle Attribute aus.
	 * @throws MeasureSerializerException
	 */
	public void asXML(XmlSerializer serializer, List<MeasureAutomatic> measureList, boolean forServer, boolean forUpdate)
			throws MeasureSerializerException {
		try {
			serializer.startTag(null, "MeasureSeries");
			serializer.attribute(null, "id", "" + id);
			if (serverId != -1) {
				serializer.attribute(null, "serverid", "" + serverId);
			}
			if (!forUpdate) {
				if (!forServer && isActive()) {
					serializer.attribute(null, "active", "true");
				}
				serializer.startTag(null, "Description");
				serializer.text(description);
				serializer.endTag(null, "Description");
				serializer.startTag(null, "Tags");
				for (String element : tags) {
					serializer.startTag(null, "Tag");
					serializer.attribute(null, "value", "" + element);
					serializer.endTag(null, "Tag");
				}
				serializer.endTag(null, "Tags");

				serializer.startTag(null, "Title");
				serializer.attribute(null, "value", title);
				serializer.endTag(null, "Title");

				if (!forServer) {
					serializer.startTag(null, "StopConditions");
					if (endTime != null) {
						serializer.startTag(null, "EndTime");
						serializer.attribute(null, "value", Util.getTimeToString(endTime));
						serializer.endTag(null, "EndTime");
					}
					if (endLocation != null) {
						serializer.startTag(null, "EndLocation");
						serializer.attribute(null, "longitude", "" + endLocation.getLongitude());
						serializer.attribute(null, "latitude", "" + endLocation.getLatitude());
						serializer.attribute(null, "accuracy", "" + endLocation.getAccuracy());
						serializer.attribute(null, "provider", "" + endLocation.getProvider());
						serializer.endTag(null, "EndLocation");
					}
					serializer.endTag(null, "StopConditions");
				}

				super.serializeAttributes(serializer, forServer);
			}

			for (MeasureAutomatic measure : measureList) {
				measure.asXML(serializer, forServer);
			}

			serializer.endTag(null, "MeasureSeries");
		} catch (Exception e) {
			throw new MeasureSerializerException(e.getMessage());
		}
	}

	protected void setAttributeFromParsingEvent(XmlPullParser parser, DataSourceIdent dataSource)
			throws MeasureParserException {
		try {
			if (parser.getEventType() == XmlPullParser.START_TAG) {
				if (parser.getName().equals("Description")) {
					description = parser.nextText();
				} else if (parser.getName().equals("Title")) {
					title = parser.getAttributeValue(null, "value");
				} else if (parser.getName().equals("EndTime")) {
					endTime = Util.getDateFromString(parser.getAttributeValue(null, "value"));
				} else if (parser.getName().equals("EndLocation")) {
					endLocation = new Location(parser.getAttributeValue(null, "provider"));
					endLocation.setLongitude(Double.parseDouble(parser.getAttributeValue(null, "longitude")));
					endLocation.setLatitude(Double.parseDouble(parser.getAttributeValue(null, "latitude")));
					endLocation.setAccuracy(Float.parseFloat(parser.getAttributeValue(null, "accuracy")));
				} else if (parser.getName().equals("Measure")) {
					if (parser.getAttributeValue(null, "type").equalsIgnoreCase("automatic")) {
						addMeasure(MeasureAutomatic.fromXML(parser, dataSource));
					} else if (parser.getAttributeValue(null, "type").equalsIgnoreCase("manual")) {
						addMeasure(MeasureManual.fromXML(parser, dataSource));
					}
				} else if (parser.getName().equals("MeasureSeries")) {
					String serverIdString = parser.getAttributeValue(null, "serverid");
					if (serverIdString != null) {
						serverId = Integer.parseInt(serverIdString);
					}
				} else if (parser.getName().equals("Tag")) {
					tags.add(parser.getAttributeValue(null, "value"));

				} else {
					super.setAttributeFromParsingEvent(parser);
				}
			}
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}
	}

	public static MeasureSeries fromXML(XmlPullParser parser) throws MeasureParserException {
		return fromXML(parser, null);
	}

	public static MeasureSeries fromXML(XmlPullParser parser, DataSourceIdent dataSource) throws MeasureParserException {
		try {
			MeasureSeries m = new MeasureSeries(Integer.parseInt(parser.getAttributeValue(null, "id")));
			m.isCommunity = dataSource != null;
			m.setRemoteDataSource(dataSource);
			
			String active_str = parser.getAttributeValue(null, "active");
			if (active_str != null && active_str.equalsIgnoreCase("true")) {
				m.setActive(true);
			}

			while (true) {
				if (parser.getName() != null) {
					m.setAttributeFromParsingEvent(parser, dataSource);
					if (parser.getEventType() == XmlPullParser.END_TAG && parser.getName().equals("MeasureSeries")) {
						break;
					}
				}
				parser.next();
			}
			if (!m.isCommunity && m.id >= nextId) {
				setNextId(m.id + 1);
			}
			return m;
		} catch (Exception e) {
			throw new MeasureParserException(e.getMessage());
		}
	}

	public void addChangeListener(MeasuresChangedListener listener) {
		if (!changedListeners.contains(listener)) {
			changedListeners.add(listener);
		}
	}

	public void removeChangeListener(MeasuresChangedListener listener) {
		changedListeners.remove(listener);
	}

	public int compareTo(MeasureAbstract another) {
		if (another instanceof MeasureSeries) {
			// Hhere Ids vor niedigere Sortieren
			return another.id - this.id;
		} else {
			// Serien vor alle anderen Messtypen sortieren
			return -1;
		}
	}

	public MeasureAutomatic findById(int id) {
		int index = Collections.binarySearch(measures, new MeasureSearchDummy(id));
		if (index >= 0) {
			return measures.get(index);
		} else {
			return null;
		}
	}

}
