package de.noisedroid.measure;

import java.io.File;
import java.nio.ShortBuffer;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.EventListener;
import java.util.EventObject;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.location.Location;
import android.location.LocationListener;
import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.MediaRecorder;
import android.os.Binder;
import android.os.Build;
import android.os.Environment;
import android.os.Handler;
import android.os.IBinder;
import android.telephony.TelephonyManager;
import de.noisedroid.NoiseDroid;
import de.noisedroid.R;
import de.noisedroid.measure.MessungsService.MessungsHelper.NoiseChangedEvent;
import de.noisedroid.measure.messmoral.MessmoralManager;
import de.noisedroid.measure.messmoral.Messmoralwertung.MoralTyp;
import de.noisedroid.util.LocationManagerHelper;
import de.noisedroid.util.Noise;

/**
 * Der grundlegende NoiseDroid-Service.
 * 
 * 
 */
public class MessungsService extends Service implements SharedPreferences.OnSharedPreferenceChangeListener {
	public enum NoiseChangeTyp {
		NEWVALUES, COMPLETED, CANCELLED
	}

	public interface NoiseChangedListener extends EventListener {
		public void noiseChanged(MessungsHelper.NoiseChangedEvent evt);

	}

	/**
	 * Klasse fr die Bindung des Services.
	 */
	public class MessungsServiceBinder extends Binder {
		public MessungsService getService() {
			return MessungsService.this;
		}
	}

	public class AudioException extends Exception {
		private static final long serialVersionUID = 1L;

		public AudioException() {
		}

		public AudioException(String detailMessage) {
			super(detailMessage);
		}
	}

	/**
	 * Hilfsklasse als {@link BroadcastReceiver} zum Empfang von Intents fr die eventbasierte Messung, sowie als
	 * {@link LocationListener} zur Durchfhrung der zeitabhngigen, automatischen Messung.
	 */
	private class AutomatischHelper extends BroadcastReceiver implements LocationManagerHelper.BetterLocationListener {

		@Override
		public void onReceive(Context context, Intent intent) {
			// Sicherstellen, dass keine Anruf luft.
			if (mTelephonyManager.getCallState() != TelephonyManager.CALL_STATE_IDLE) {
				return;
			}

			Location location = mLocationManagerHelper.getBestLastKnownLocation();
			if (location != null) {
				// TODO Wichtig: Neue aktuelle Position anfordern
				// long t = location.getTime();
				// long t2 = System.currentTimeMillis();
				// long timeNowDelta = t2 - t;
				// if (timeNowDelta < 30000) {
				automatischMessen(location);
				// }
			}
		}

		public void onBetterLocation(Location location) {
			// Abbrechen, wenn nicht mindestens die Hlfte des anzustrebenden Intervalls ablief
			if ((System.currentTimeMillis() - lastUpdate) < updateInterval / 2) {
				return;
			}

			// Sicherstellen, dass keine Anruf luft.
			if (mTelephonyManager.getCallState() != TelephonyManager.CALL_STATE_IDLE) {
				return;
			}

			automatischMessen(location);
		}

		private void automatischMessen(final Location location) {
			// Messmoral mit einem Dummyobjekt an den Start bringen, um die Moral spter sicher wieder freizugeben ohne
			// dem eigentlichen Messsystem in die Quere zu kommen.
			final Object dummy = new Object();
			mMessmoralManager.start(dummy);

			// Messung 2 Sek. verzgern um Messmoral zu prfen.
			Handler handler = new Handler();
			handler.postDelayed(new Runnable() {
				public void run() {
					if (mMessmoralManager.getMoralWert() >= 50) {
						lastUpdate = System.currentTimeMillis(); // Nur bei potenziell nutzbaren Messungen die Zeit
																	// aktualisieren, erlaubt schnellere Wiederholung
						final MeasureAutomatic messung = new MeasureAutomatic();
						messung.setLocation(location);

						MessungsHelper helper = getMessung(5000);
						helper.addNoiseChangedListener(new NoiseChangedListener() {

							public void noiseChanged(NoiseChangedEvent evt) {
								if (evt.getTyp() == NoiseChangeTyp.COMPLETED) {
									// TODO Man knnt noch von der mittleren Moral whrend der Messuung abhngig machen,
									// ob die
									// Messung brauchbar ist
									messung.setNoise(evt.getMessungsHelper().getAvgNoise());
									messung.setMoral(evt.getMessungsHelper().getAvgMoral());
									Calendar c = new GregorianCalendar();

									messung.setTime(c);
									MeasureManager.getInstance().addMeasure(messung);
								}
							}
						});
						try {
							helper.start();
						} catch (AudioException e) {

						}
					}

					// Messmoral mit Dummyobjekt freigeben, damit keine Probleme auftauchen wenn Messmoral woanders
					// genutzt wird.
					mMessmoralManager.stop(dummy);
				}
			}, 2000);
		}

	}

	/**
	 * Hilfsklasse zur Reprsentation eines Messvorganges. Wird von {@link MessungsService} instanziiert und mit Daten
	 * vom Mikrofon versorgt. Somit sind auch, falls vorkommend, verschiedene parallele Messungen mglich.
	 */
	public class MessungsHelper {

		/**
		 * Event zum Signalisieren neuer Messwerte bzw. Messeigenschaften whrend eines Messvorganges.
		 */
		public class NoiseChangedEvent extends EventObject {
			private static final long serialVersionUID = 1L;
			private NoiseChangeTyp typ;

			public NoiseChangedEvent(MessungsHelper source, NoiseChangeTyp typ) {
				super(source);
				this.typ = typ;
			}

			public NoiseChangeTyp getTyp() {
				return typ;
			}

			public MessungsHelper getMessungsHelper() {
				return (MessungsHelper) source;
			}
		}

		private float minNoise;
		private float maxNoise;
		private float lastNoise;
		private float avgNoise;
		private float lastMoral;
		private float avgMoral;
		private long sumPCM;
		private int anzPCM;
		private int sumMoral;
		private int anzMoral;
		private float fortschritt;
		private boolean completed = false;
		private boolean stopped = true;
		// Wird hauptschlich genutzt, falls in Zukunft Zugriff auf jeden PCM-Wert
		// (und nicht nur Mittelwerte) ntig ist, z. B. fr Frequenzgewichtung.
		private ShortBuffer gesamtBuffer;
		private ArrayList<NoiseChangedListener> noiseChangedListeners = new ArrayList<NoiseChangedListener>();
		// Handler um die Events auf dem UI-Thread an die Listener zu leiten.
		private Handler mainHandler = new Handler(getMainLooper());

		/**
		 * Konstruktor
		 * 
		 * @param bufferSize
		 *            Gre des zu fllenden Buffers, definiert Lnge Messung
		 */
		private MessungsHelper(int bufferSize) {
			gesamtBuffer = ShortBuffer.allocate(bufferSize);
		}

		public void addNoiseChangedListener(NoiseChangedListener listener) {
			noiseChangedListeners.add(listener);
		}

		public void removeNoiseChangedListener(NoiseChangedListener listener) {
			noiseChangedListeners.remove(listener);
		}

		/**
		 * Wird vom {@link MessungsService} aufgerufen wenn neue Mikrofon-PCM-Daten verfgbar wurden.
		 * 
		 * @param pcm
		 *            Array der Werte
		 * @param anz
		 *            Anzahl der darin zu verwendenden Indizes, ab Index 0
		 */
		public void newPCM(short[] pcm, int anz) {
			// Festlegung der von anz noch ntigen Anzahl neuer Werte (ansonsten folgt Overflow)
			anz = Math.min(gesamtBuffer.remaining(), anz);
			gesamtBuffer.put(pcm, 0, anz);
			anzPCM += anz;

			int moral = mMessmoralManager.getMoralWert();

			// Summieren der Moral
			sumMoral += moral;
			anzMoral++;

			// Berechnung des Durchschnittswertes der neuen Werte
			long tempSumPCM = 0;
			for (int i = 0; i < anz; i++) {
				int squarePCM = pcm[i] * pcm[i];
				sumPCM += squarePCM;
				tempSumPCM += squarePCM;
			}
			lastNoise = Noise.pcmToDB((float) Math.sqrt(tempSumPCM / (anz * 1f)));
			lastMoral = moral;

			// Berechnung des (bisherigen) Gesamtdurchschnitts
			avgNoise = Noise.pcmToDB((float) Math.sqrt(sumPCM / (anzPCM * 1f)));
			avgMoral = sumMoral / (anzMoral * 1f);

			// Setzen der Extremwerte
			if (lastNoise > maxNoise) {
				maxNoise = lastNoise;
			}
			if (lastNoise < minNoise) {
				minNoise = lastNoise;
			}

			fortschritt = anzPCM / (float) gesamtBuffer.capacity();

			if (gesamtBuffer.remaining() == 0) {
				// Falls geforderte Anzahl an Werten erreicht wurde, werden keine weiteren vom Service gefordert
				removeMessungsHelper(this);
				noiseChanged(NoiseChangeTyp.COMPLETED);
			} else {
				noiseChanged(NoiseChangeTyp.NEWVALUES);
			}
		}

		public float getAvgNoise() {
			return avgNoise;
		}

		public float getLastNoise() {
			return lastNoise;
		}

		public float getMinNoise() {
			if (minNoise == Float.MAX_VALUE) {
				return 0;
			} else {
				return minNoise;
			}
		}

		public float getMaxNoise() {
			return maxNoise;
		}

		public float getAvgMoral() {
			return avgMoral;
		}

		public float getLastMoral() {
			return lastMoral;
		}

		public float getFortschritt() {
			return fortschritt;
		}

		public List<MoralTyp> getMoralTypen() {
			return mMessmoralManager.getMoralTyp();
		}

		/*
		 * http://code.google.com/p/moonblink/source/browse/trunk/HermitLibrary/src/org/hermit/dsp/SignalPower.java
		 * 
		 * public final static double calculatePowerDb(short[] sdata, int off, int samples) { // Calculate the sum of
		 * the values, and the sum of the squared values. // We need longs to avoid running out of bits. double sum = 0;
		 * double sqsum = 0; for (int i = 0; i < samples; i++) { final long v = sdata[off + i]; sum += v; sqsum += v *
		 * v; }
		 * 
		 * // sqsum is the sum of all (signal+bias), so // sqsum = sum(signal) + samples * bias // hence //
		 * sum(signal) = sqsum - samples * bias // Bias is simply the average value, i.e. // bias = sum / samples //
		 * Since power = sum(signal) / samples, we have // power = (sqsum - samples * sum / samples) / samples // so
		 * // power = (sqsum - sum / samples) / samples double power = (sqsum - sum * sum / samples) / samples;
		 * 
		 * // Scale to the range 0 - 1. power /= MAX_16_BIT * MAX_16_BIT;
		 * 
		 * // Convert to dB, with 0 being max power. Add a fudge factor to make // a "real" fully saturated input come
		 * to 0 dB. return Math.log10(power) * 10f + FUDGE; }
		 */

		/**
		 * Informiert Listener ber neue Werte, stellt Ausfhrung auf dem richtigen Thread sicher.
		 * 
		 * @param typ
		 *            Typ des Ereignisses
		 */
		private void noiseChanged(final NoiseChangeTyp typ) {
			if (stopped) {
				return;
			}
			if (typ == NoiseChangeTyp.COMPLETED) {
				completed = true;
			}
			mainHandler.post(new Runnable() {
				public void run() {
					NoiseChangedEvent e = new NoiseChangedEvent(MessungsHelper.this, typ);
					for (NoiseChangedListener listener : noiseChangedListeners) {
						listener.noiseChanged(e);
					}
				}
			});
		}

		public boolean isCompleted() {
			return completed;
		}

		/**
		 * Beendet aktuelle Messung. Falls eine Mindestzahl an Mikrofon-Werten erreicht wurde gilt Messung als
		 * erfolgreich.
		 */
		public void stop() {
			removeMessungsHelper(this);
			stopped = true;
			if (!completed) {
				if (anzPCM >= 2000) {
					noiseChanged(NoiseChangeTyp.COMPLETED);
				} else {
					noiseChanged(NoiseChangeTyp.CANCELLED);
				}
			}
		}

		/**
		 * Startet die Messung (neu), setzt Variablen zurck und fordert neue Werte beim {@link MessungsService} an
		 * 
		 * @throws AudioException
		 */
		public void start() throws AudioException {
			sumPCM = anzPCM = sumMoral = anzMoral = 0;
			avgNoise = maxNoise = lastNoise = avgMoral = lastMoral = fortschritt = 0;
			minNoise = Float.MAX_VALUE;
			completed = false;
			gesamtBuffer.clear();
			stopped = false;
			addMessungsHelper(this);
		}
	}

	private NotificationManager mNotificationManager;
	private LocationManagerHelper mLocationManagerHelper;
	private MessmoralManager mMessmoralManager;
	private TelephonyManager mTelephonyManager;
	private final IBinder mBinder = new MessungsServiceBinder(); // Bindungsobjekt zum Nutzen des Services

	// Audioressourcen und Einstellungen
	private AudioRecord recorder;
	private int minBufferSize;

	// Folgende Werte werden durch die audio.xml berschrieben
	private int channel = AudioFormat.CHANNEL_IN_MONO;
	private int format = AudioFormat.ENCODING_PCM_16BIT;
	private int rate = 11025;
	private int source = MediaRecorder.AudioSource.DEFAULT;
	private int measureConfVersion;

	// Liste der mit Audioaten "zu versorgenden" Messungshelfer, CopyOnWriteArrayList stellt Threadsicherheit sicher,
	// Geschwindigkeitsverlust kann hier vernachlssigt werden.
	private List<MessungsHelper> messungsHelperList = new CopyOnWriteArrayList<MessungsHelper>();

	private boolean eventMeasures = false;

	private Thread aufnahmeThread;
	private AutomatischHelper mAutomatischHelper = new AutomatischHelper();
	private SharedPreferences mSharedPreferences;

	// Um anzuzeigen, dass der Service noch ntig ist...
	private boolean clientsConnected = false;

	// Intervall und lezten Zeitpunkt fr automatische Messung zwischenspeichern um zu hufige Messungen zu vermeiden.
	private int updateInterval;
	private long lastUpdate;

	private Object stopSync = new Object();

	@Override
	public void onCreate() {
		mNotificationManager = (NotificationManager) getSystemService(NOTIFICATION_SERVICE);
		loadAudioSettings();
		mMessmoralManager = new MessmoralManager(this);
		mLocationManagerHelper = new LocationManagerHelper(this);
		mLocationManagerHelper.addBetterLocationListener(mAutomatischHelper);
		mTelephonyManager = (TelephonyManager) getSystemService(Context.TELEPHONY_SERVICE);

		// Zugriff auf die App-Einstellungen
		mSharedPreferences = getSharedPreferences("preferences", MODE_PRIVATE);
		mSharedPreferences.registerOnSharedPreferenceChangeListener(this);

		setBenachrichtigung(null, "");

		setUpLocationUpdates(); // Initialisieren der ntigen Location-Updates
		setMeasureEvents();
	}

	private void setMeasureEvents() {
		eventMeasures = false;
		try {
			getApplicationContext().unregisterReceiver(mAutomatischHelper);
		} catch (Exception e) {
			// Gibt halt keine Methode um zu prfen ob der schon registriert war
		}
		if (mSharedPreferences.getBoolean("eventMessungAktiv", false)) {
			String events = mSharedPreferences.getString("events", null);
			if (events != null && events.length() != 0) {
				IntentFilter filter = new IntentFilter();
				String[] eventArray = events.split(";");
				for (String event : eventArray) {
					filter.addAction(event);
				}
				getApplicationContext().registerReceiver(mAutomatischHelper, filter);
				eventMeasures = true;
			}
		}
	}

	/**
	 * Ldt die fr dieses Gert bestimmten Audioeinstellungen
	 */
	public void loadAudioSettings() {
		try {
			String audioPath = Environment.getExternalStorageDirectory()
					+ "/Android/data/de.noisedroid/xml/measureconf.xml";
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			Document dom;
			if (new File(audioPath).exists()) {
				DocumentBuilder builder = factory.newDocumentBuilder();
				dom = builder.parse(new File(audioPath));
			} else {
				DocumentBuilder builder = factory.newDocumentBuilder();
				dom = builder.parse(getResources().openRawResource(R.raw.measureconf));
			}

			// Version auslesen
			measureConfVersion = Integer.parseInt(dom.getDocumentElement().getAttribute("version"));

			// Suche des zum Gerte passenden Settings-Knoten
			Element settingsNode = null;
			NodeList settingsNodes = dom.getElementsByTagName("Settings");
			for (int i = 0; i < settingsNodes.getLength(); i++) {
				String attrWert = settingsNodes.item(i).getAttributes().getNamedItem("product").getNodeValue();
				if (attrWert.equals(Build.PRODUCT) || attrWert.equals("general")) {
					settingsNode = (Element) settingsNodes.item(i);
					break;
				}
			}

			rate = Integer.parseInt(settingsNode.getElementsByTagName("Rate").item(0).getAttributes()
					.getNamedItem("value").getNodeValue());
			channel = Integer.parseInt(settingsNode.getElementsByTagName("Channel").item(0).getAttributes()
					.getNamedItem("value").getNodeValue());
			format = Integer.parseInt(settingsNode.getElementsByTagName("Format").item(0).getAttributes()
					.getNamedItem("value").getNodeValue());
			source = Integer.parseInt(settingsNode.getElementsByTagName("Source").item(0).getAttributes()
					.getNamedItem("value").getNodeValue());

			minBufferSize = AudioRecord.getMinBufferSize(rate, channel, format);
			if (minBufferSize == AudioRecord.ERROR_BAD_VALUE || minBufferSize == AudioRecord.ERROR) {
				// TODO groer Fehler
			}
		} catch (Exception e) {

		}
	}

	/**
	 * Gibt die aktuell verwendete Versionsnummer der measureconf.xml zurck
	 * 
	 * @return
	 */
	public int getMeasureConfVersion() {
		return measureConfVersion;
	}

	/**
	 * Registriert den {@link AutomatischHelper} fr die ntigen Location-Updates, basierend auf den App-Einstellungen
	 */
	private void setUpLocationUpdates() {
		// Zunchst alle Registrierungen zurcknehmen...
		mLocationManagerHelper.stop();

		// Und danach neu setzen
		if (mSharedPreferences.getBoolean("autoMessungAktiv", false)) {
			updateInterval = Integer.parseInt(mSharedPreferences.getString("autoMessungIntervall", "120000"));
			mLocationManagerHelper.findBestLocation(updateInterval, 0);
		}

		// Service beenden, falls nicht mehr bentigt
		ntigkeitPrfen();
	}

	/**
	 * Bereitet die Audioressourcen zum Messen vor und benachrichtigt den Nutzer.
	 * 
	 * @throws AudioException
	 */
	private void setUpRecorder() throws AudioException {
		synchronized (stopSync) {
			if (recorder == null || recorder.getState() == AudioRecord.STATE_UNINITIALIZED) {
				recorder = new AudioRecord(source, rate, channel, format, minBufferSize * 2);
				if (recorder.getState() == AudioRecord.STATE_UNINITIALIZED) {
					throw new AudioException();
				}
			}
			if (recorder.getRecordingState() == AudioRecord.RECORDSTATE_STOPPED) {
				recorder.startRecording();
				setBenachrichtigung(getString(R.string.l_rmmessung_), getString(R.string.l_rmmessung_l_uft_));
			}
			mMessmoralManager.start(this);
		}
	}

	/**
	 * Gibt die Audioressourcen frei und benachrichtigt den Nutzer.
	 */
	private void stopRecorder() {
		synchronized (stopSync) {
			if (recorder != null) {
				if (recorder.getState() != AudioRecord.STATE_UNINITIALIZED) {
					recorder.stop();
				}
				recorder.release();
				recorder = null;
				setBenachrichtigung(getString(R.string.messung_beendet), "");
			}
			mMessmoralManager.stop(this);
		}
	}

	/**
	 * Verwenden um Audiodaten anzufordern. Fhrt das Sammeln und verteilen der Audiodaten in einem separaten Thread
	 * durch, mhermaliges Aufrufen trotz Aufnahme ist unproblematisch
	 * 
	 * @throws AudioException
	 */
	private void aufnehmen() throws AudioException {
		if (!nimmtAuf()) {
			// Neuer Thread nur falls er noch nicht existiert oder wieder beendet wurde.
			aufnahmeThread = new Thread() {
				@Override
				public void run() {
					aufnahme();
				}
			};
			setUpRecorder();
			aufnahmeThread.start();
		}
	}

	/**
	 * Methode fragt die Audioressourcen ab und verteilt die Daten an alle {@link MessungsHelper}. Nachdem keine
	 * weiteren Daten mehr bentigt werden, werden die Audioressourcen wieder friegegeben. Sollte nicht direkt
	 * aufgerufen werden!
	 */
	private void aufnahme() {
		// TODO Messabbruch bei Anruf
		short[] tempBuffer = new short[minBufferSize];
		while (messungsHelperList.size() != 0) {
			if (recorder == null) {
				break;
			}
			int anz = recorder.read(tempBuffer, 0, minBufferSize);

			for (MessungsHelper helper : messungsHelperList) {
				helper.newPCM(tempBuffer, anz);
			}

		}
		if (messungsHelperList.size() == 0) {
			stopRecorder();
		}
	}

	/**
	 * ...startet automatisch das Aufnehmen
	 * 
	 * @param helper
	 * @throws AudioException
	 */
	private void addMessungsHelper(MessungsHelper helper) throws AudioException {
		messungsHelperList.add(helper);
		if (messungsHelperList.size() != 0 && !nimmtAuf()) {
			aufnehmen();
		}
	}

	private boolean nimmtAuf() {
		return (aufnahmeThread != null && aufnahmeThread.isAlive());
	}

	private void removeMessungsHelper(MessungsHelper helper) {
		messungsHelperList.remove(helper);
	}

	private void setBenachrichtigung(String ticker, String text) {
		Notification notification = new Notification(R.drawable.statusicon, ticker, 0);
		Intent notificationIntent = new Intent(this, NoiseDroid.class);
		PendingIntent contentIntent = PendingIntent.getActivity(this, 0, notificationIntent, 0);

		notification.setLatestEventInfo(getApplicationContext(), getString(R.string.noisedroid_l_rmmessdienst), text, contentIntent);
		notification.flags |= Notification.FLAG_ONGOING_EVENT;

		// notification.number
		mNotificationManager.notify(1, notification);
	}

	@Override
	public int onStartCommand(Intent intent, int flags, int startId) {
		return START_STICKY;
	}

	@Override
	public IBinder onBind(Intent arg0) {
		return mBinder;
	}

	@Override
	public void onDestroy() {
		stopRecorder();
		mNotificationManager.cancel(1);
		mLocationManagerHelper.stop();
		try {
			getApplicationContext().unregisterReceiver(mAutomatischHelper);
		} catch (Exception e) {
			// Gibt halt keine Methode um zu prfen ob der schon registriert war
		}
		super.onDestroy();
	}

	@Override
	public boolean onUnbind(Intent intent) {
		mNotificationManager.cancel(1);
		stopRecorder();
		clientsConnected = false;
		ntigkeitPrfen();
		return true;
	}

	@Override
	public void onRebind(Intent intent) {
		clientsConnected = true;
		super.onRebind(intent);
	}

	/**
	 * Methode um eine Messung vorgegebener Messdauer durchzufhren. Gibt {@link MessungsHelper} zurck, um ber den
	 * Messverlauf informiert zu werden.
	 * 
	 * @param dauer
	 * @return
	 */
	public MessungsHelper getMessung(int dauer) {
		return new MessungsHelper((int) (rate * (dauer / 1000f)));
	}

	public void onSharedPreferenceChanged(SharedPreferences sharedPreferences, String key) {
		// Wohl unntig? key.equals("autoProviderAktiv") || key.equals("locationProviders") ||
		if (key.equals("autoMessungIntervall") || key.equals("autoMessungAktiv")) {
			// Die Location-Updates nochmal berdenken, falls Einstellungen diesbezglich gendert wurden.
			setUpLocationUpdates();
		} else if (key.equals("eventMessungAktiv") || key.equals("events")) {
			setMeasureEvents();
		}
	}

	/**
	 * Stoppt den Service, falls er unntig wurde (kein Client, keine automatische Messung, ...)
	 */
	private void ntigkeitPrfen() {
		if (!clientsConnected && !mLocationManagerHelper.isAktiv() && !eventMeasures) {
			stopSelf();
		}

	}
}
