package de.noisedroid.measure.messmoral;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import android.content.Context;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.os.Build;
import android.os.Environment;
import de.noisedroid.R;
import de.noisedroid.measure.messmoral.Messmoralwertung.MoralTyp;

/**
 * Der MessmoralManager ldt alle fr die Bestimmung der Messmoral auf dem aktuellen mobilen Gert ntigen Daten aus der
 * raw/moralconf.xml (MML genannt) und sammelt kontinuierlich die entsprechenden Sensordaten. Er erlaubt jederzeit
 * einfachen Zugriff auf die aktuelle Moral. Die Moral ist ein simpler Indikator fr die Qualitt und Relevanz eines
 * Messvorhabens und sollte deshalb stets bei Messungen herangezogen werden.
 */
public class MessmoralManager implements SensorEventListener {
	private List<Messmoralwertung> moralChangeListe = new ArrayList<Messmoralwertung>(4);
	// HashMap-Struktur welche fr jeden Sensor die Referenz auf die aktuellen Sensordaten enthlt.
	private Map<String, SensordataReference<float[]>> sensordataMap = new HashMap<String, SensordataReference<float[]>>();
	private int moralConfVersion;
	private Context context;
	private SensorManager mSensorManager;
	private List<Sensor> ntigeSensoren = new ArrayList<Sensor>();
	private boolean started = false;
	private List<Object> moraluserList = new ArrayList<Object>();

	public MessmoralManager(Context context) {
		this.context = context;
		mSensorManager = (SensorManager) context.getSystemService(Context.SENSOR_SERVICE);

		loadMoralSettings();
	}

	/**
	 * Ldt die Moral fr dieses Gert
	 */
	public void loadMoralSettings() {
		try {
			String moralPath = Environment.getExternalStorageDirectory()
					+ "/Android/data/de.noisedroid/xml/moralconf.xml";
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			Document dom;
			if (new File(moralPath).exists()) {
				DocumentBuilder builder = factory.newDocumentBuilder();
				dom = builder.parse(new File(moralPath));
			} else {
				DocumentBuilder builder = factory.newDocumentBuilder();
				dom = builder.parse(context.getResources().openRawResource(R.raw.moralconf));
			}

			// Version auslesen
			moralConfVersion = Integer.parseInt(dom.getDocumentElement().getAttribute("version"));

			Element messmoralNode = null;
			NodeList messmoralNodes = dom.getElementsByTagName("Messmoral");
			for (int i = 0; i < messmoralNodes.getLength(); i++) {
				String attrWert = messmoralNodes.item(i).getAttributes().getNamedItem("product").getNodeValue();
				if (attrWert.equals(Build.PRODUCT) || attrWert.equals("general")) {
					messmoralNode = (Element) messmoralNodes.item(i);
					break;
				}
			}

			moralChangeListe.clear(); // Liste der Moralnderungen zurcksetzen
			NodeList moralnderungNodes = messmoralNode.getElementsByTagName("Moralchange");
			for (int i = 0; i < moralnderungNodes.getLength(); i++) {
				Messmoralwertung tMessmoralChange = new Messmoralwertung(moralnderungNodes.item(i), this);
				moralChangeListe.add(tMessmoralChange);
			}

			// Ntige Sensoren finden
			List<String> ntigeSensorenNamen = getSensorsInMessmoral(messmoralNode);
			ntigeSensoren.clear();
			// Sensorobjekte fr Sensornamen finden
			List<Sensor> sensors = mSensorManager.getSensorList(Sensor.TYPE_ALL);
			for (Sensor sensor : sensors) {
				if (ntigeSensorenNamen.contains(sensor.getName())) {
					ntigeSensoren.add(sensor);
				}
			}
		} catch (Exception e) {

		}
	}

	private void registerSensors() {
		mSensorManager.unregisterListener(this); // MessmoralManager von allen Sensoren abmelden
		for (Sensor sensor : ntigeSensoren) {
			mSensorManager.registerListener(this, sensor, SensorManager.SENSOR_DELAY_NORMAL);
		}
	}

	private void unregisterSensors() {
		mSensorManager.unregisterListener(this);
	}

	public void start(Object caller) {
		if (!moraluserList.contains(caller)) {
			moraluserList.add(caller);
		}
		if (!started) {
			registerSensors();
			started = true;
		}
	}

	public void stop(Object caller) {
		moraluserList.remove(caller);

		if (moraluserList.size() == 0) {
			unregisterSensors();
			for (SensordataReference<float[]> dataRef : sensordataMap.values()) {
				dataRef.valid = false;
			}
			sensordataMap.clear();
			started = false;
		}
	}

	public int getMoralConfVersion() {
		return moralConfVersion;
	}

	public String getSensorNameByType(int type) {
		if (mSensorManager.getSensorList(type).size() == 0) {
			return null;
		} else {
			return mSensorManager.getSensorList(type).get(0).getName();
		}
	}

	/**
	 * Gibt Liste aller von diesem MML-Abschnitt verwendeten Sensornamen zurck.
	 * 
	 * @param node
	 *            Messmoral-Tag
	 * @return
	 * @throws MoralException
	 *             Falls MML fehlerhaft ist
	 */
	public List<String> getSensorsInMessmoral(Element node) throws MoralException {
		Vector<String> sensorNames = new Vector<String>();
		try {
			NodeList nodes = node.getElementsByTagName("Sensor");

			for (int i = 0; i < nodes.getLength(); i++) {
				String name;
				if (nodes.item(i).getAttributes().getNamedItem("name") != null) {
					name = nodes.item(i).getAttributes().getNamedItem("name").getNodeValue();
				} else {
					int type = Integer.parseInt(nodes.item(i).getAttributes().getNamedItem("type").getNodeValue());
					name = getSensorNameByType(type);
				}
				if (name != null && !sensorNames.contains(name)) {
					sensorNames.add(name);
				}
			}
		} catch (Exception e) {
			throw new MoralException();
		}
		return sensorNames;
	}

	public void onAccuracyChanged(Sensor sensor, int accuracy) {
		// TODO Auto-generated method stub

	}

	public void onSensorChanged(SensorEvent event) {
		// Lookup des zuletzt gemessenen Wertes fr diesen Sensor
		SensordataReference<float[]> oldValues = sensordataMap.get(event.sensor.getName());
		if (oldValues != null) {
			// Behandlung der speziellen "Diff" Sensortypen, Berechnung der Sensorwertdifferenzen
			switch (event.sensor.getType()) {
			case Sensor.TYPE_ACCELEROMETER:
			case Sensor.TYPE_GYROSCOPE:
				float[] diffValues = new float[oldValues.data.length];
				for (int i = 0; i < oldValues.data.length; i++) {
					diffValues[i] = Math.abs(event.values[i] - oldValues.data[i]);
				}
				// Setzen der neuen Differenz-Sensordaten
				putSensorData(event.sensor.getName() + " Diff", diffValues);
				break;
			}
		}

		// Setzen der neuen Sensordaten
		putSensorData(event.sensor.getName(), event.values.clone());
	}

	/**
	 * Speichert Sensorwerte fr Sensor. "Wrapt" die Daten in spezieller Struktur fr schnellen Zugriff, falls noch
	 * nicht geschehen.
	 * 
	 * @param key
	 *            Sensorname
	 * @param values
	 *            Sensordaten
	 */
	private void putSensorData(String key, float[] values) {
		SensordataReference<float[]> oldValues = sensordataMap.get(key);
		if (oldValues == null) {
			sensordataMap.put(key, new SensordataReference<float[]>(values));
		} else {
			oldValues.data = values;
		}
	}

	/**
	 * Gibt die "Referenz" auf die aktuellsten Sensordaten fr bestimmten Sensornamen zurck. Das zurckgegebene Objekt
	 * wird kontinuierlich mit den neusten Sensordaten aktualisiert und muss nicht neu abgefragt werden.
	 * 
	 * @param name
	 *            Sensorname
	 * @return Sensordaten Objekt welches die aktuellsten Daten reprsentiert
	 * @throws MoralException
	 *             Falls fr diesen Sensornamen (noch) keine Daten vorliegen
	 */
	public SensordataReference<float[]> getSensorData(String name) throws MoralException {
		SensordataReference<float[]> values = sensordataMap.get(name);
		if (values != null) {
			return values;
		} else {
			throw new MoralException();
		}
	}

	/**
	 * Berechnet den Moralwert aus den aktuellsten Sensordaten. Beziehen sich Moralteilauswertungen auf (noch) nicht
	 * vorhandene Sensordaten, werden sie missachtet.
	 * 
	 * @return Aktueller Moralwert
	 */
	public int getMoralWert() {
		int moralwert = 100;
		for (Messmoralwertung messmoralChange : moralChangeListe) {
			try {
				if (messmoralChange.auswerten()) {
					moralwert += messmoralChange.getMoralwert();
				}
			} catch (MoralException e) {

			}
		}
		return moralwert;
	}

	public List<MoralTyp> getMoralTyp() {
		ArrayList<MoralTyp> typen = new ArrayList<Messmoralwertung.MoralTyp>();
		for (Messmoralwertung messmoralChange : moralChangeListe) {
			if (messmoralChange.getTyp() == MoralTyp.UNSPECIFIED) {
				continue;
			}
			try {
				Boolean result = messmoralChange.auswerten();
				int moralwert = messmoralChange.getMoralwert();
				// Ereignistyp gilt als eingetreten falls
				// Moralwertung positiv ausgewertet und Moralnderung negativ bzw.
				// Moralwertung negativ ausgewertet aber Moralnderung positiv wre
				if ((result && moralwert < 0) || (!result && moralwert > 0)) {
					typen.add(messmoralChange.getTyp());
				}
			} catch (MoralException e) {

			}
		}
		return typen;
	}

}
