package de.noisedroid.osm;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Cap;
import android.graphics.Rect;
import android.graphics.RectF;
import android.location.Location;
import android.util.AttributeSet;
import android.view.View;

import com.google.android.maps.GeoPoint;

/**
 * Eine einfache View die basierend auf den heruntergeladenen OSM-Daten eine Kartenansicht bietet. Nur fr kleine
 * Ausschnitte geeignet
 * 
 */
public class StaticMapView extends View {
	private List<GeoPoint> punktListe;
	private TileInfo tileInfo;
	private RectF punktExtent = new RectF();
	private Matrix transform = new Matrix();
	private float[] punkteSrc;
	private float[] punkteDst;
	private float[] punktCenter = new float[2];
	private Paint paint;
	private int maxZoom = 15;
	private int centerX = -1;
	private int centerY = -1;

	public StaticMapView(Context context) {
		super(context);
		init();
	}

	public StaticMapView(Context context, AttributeSet attrs) {
		super(context, attrs);
		maxZoom = attrs.getAttributeIntValue(null, "maxZoom", 16);
		centerX = attrs.getAttributeIntValue(null, "centerX", -1);
		centerY = attrs.getAttributeIntValue(null, "centerY", -1);
		float lat = attrs.getAttributeFloatValue(null, "lat", -1);
		float lon = attrs.getAttributeFloatValue(null, "lon", -1);
		if (lat != -1 && lon != -1) {
			GeoPoint p = new GeoPoint((int) (lat * 1E6), (int) (lon * 1E6));
			setPunkt(p);
		}
		init();
	}

	public StaticMapView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		init();
	}

	private void init() {
		paint = new Paint();
		paint.setAntiAlias(true);
		paint.setStrokeCap(Cap.ROUND);
		paint.setStrokeWidth(10);
		paint.setColor(Color.BLUE);
	}

	public void setMaxZoom(int maxZoom) {
		this.maxZoom = maxZoom;
		setTileInfo();
		setTransform();
	}

	public void setPunktListe(List<GeoPoint> punktListe) {
		this.punktListe = new ArrayList<GeoPoint>(punktListe);
		punkteSrc = new float[punktListe.size() * 2];
		punkteDst = new float[punktListe.size() * 2];
		if (punktListe.size() != 0) {
			punktExtent.left = punktExtent.right = punktListe.get(0).getLongitudeE6();
			punktExtent.bottom = punktExtent.top = punktListe.get(0).getLatitudeE6();
		} else {
			punktExtent.setEmpty();
		}
		for (int i = 0; i < punktListe.size(); i++) {
			GeoPoint p = punktListe.get(i);
			punktExtent.union(p.getLongitudeE6(), p.getLatitudeE6());
			punkteSrc[2 * i] = p.getLongitudeE6();
			punkteSrc[2 * i + 1] = p.getLatitudeE6();
		}
		setTileInfo();
		setTransform();
		invalidate();
	}

	public void setPunktListeLocation(List<Location> locationListe) {
		List<GeoPoint> punkte = new ArrayList<GeoPoint>();
		for (Location l : locationListe) {
			punkte.add(new GeoPoint((int) (l.getLatitude() * 1E6), (int) (l.getLongitude() * 1E6)));
		}
		setPunktListe(punkte);
	}

	public void setPunkt(Location location) {
		GeoPoint p = new GeoPoint((int) (location.getLatitude() * 1E6), (int) (location.getLongitude() * 1E6));
		setPunkt(p);
	}

	public void setPunkt(GeoPoint punkt) {
		punktListe = new ArrayList<GeoPoint>();
		punktListe.add(punkt);
		punktExtent.left = punktExtent.right = punkt.getLongitudeE6();
		punktExtent.bottom = punktExtent.top = punkt.getLatitudeE6();
		punkteSrc = new float[2];
		punkteDst = new float[2];
		punkteSrc[0] = punkt.getLongitudeE6();
		punkteSrc[1] = punkt.getLatitudeE6();
		setTileInfo();
		setTransform();
		invalidate();
	}

	private void setTransform() {
		if (tileInfo == null)
			return;

		Rect srcE6 = tileInfo.getLLBBoxE6();

		int dX = srcE6.width();
		int dY = srcE6.height();

		transform.reset();
		transform.preScale(256f / dX, 256f / dY);
		transform.preTranslate(-srcE6.left, -srcE6.top);

		transform.mapPoints(punkteDst, punkteSrc);

		transform.mapPoints(punktCenter, new float[] { punktExtent.centerX(), punktExtent.centerY() });
	}

	private void setTileInfo() {
		if (punktListe == null || punktListe.size() == 0) {
			tileInfo = null;
			return;
		}

		TileInfo resultTileInfo = null;
		int tileX = -1;
		int tileY = -1;
		int zoom = maxZoom;
		Iterator<GeoPoint> iter = punktListe.iterator();
		GeoPoint p = iter.next();
		while (p != null) {
			if (zoom < 0) {
				tileInfo = null;
				return;
			}
			resultTileInfo = TileInfo.getTileInfoForLonLat(p, zoom);
			if (tileX == -1 || tileY == -1) {
				tileX = resultTileInfo.x;
				tileY = resultTileInfo.y;
			} else if (tileX != resultTileInfo.x || tileY != resultTileInfo.y
					|| !TileManager.tileExists(resultTileInfo)) {
				tileX /= 2;
				tileY /= 2;
				zoom--;
			} else {
				if (iter.hasNext())
					p = iter.next();
				else
					p = null;
			}
		}

		tileInfo = resultTileInfo;
	}

	@Override
	protected void onDraw(Canvas canvas) {
		super.onDraw(canvas);
		if (tileInfo == null || canvas.getClipBounds().isEmpty()) {
			return;
		}

		int dx, dy;
		if (centerX == -1) {
			dx = (int) (getWidth() / 2f - punktCenter[0]);
		} else {
			dx = (int) (centerX - punktCenter[0]);
		}

		if (centerY == -1) {
			dy = (int) (getHeight() / 2f - punktCenter[1]);
		} else {
			dy = (int) (centerY - punktCenter[1]);
		}

		canvas.clipRect(new Rect(getPaddingLeft(), getPaddingTop(), getWidth() - getPaddingRight(), getHeight()
				- getPaddingBottom()));
		canvas.translate(dx, dy);
		Rect clip = canvas.getClipBounds();

		for (int y = -1; y <= 1; y++) {
			if ((y + 1) * 256 <= clip.top || y * 256 >= clip.bottom) {
				continue;
			}
			for (int x = -1; x <= 1; x++) {
				if ((x + 1) * 256 <= clip.left || x * 256 >= clip.right) {
					continue;
				}
				TileInfo t = new TileInfo(tileInfo.x + x, tileInfo.y + y, tileInfo.zoom);
				Bitmap tile = TileManager.getTile(t);
				if (tile != null) {
					canvas.drawBitmap(TileManager.getTile(t), x * 256, y * 256, null);
				}
			}
		}

		canvas.drawPoints(punkteDst, paint);
	}

}
