package de.noisedroid.osm;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.concurrent.ArrayBlockingQueue;

import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.OnSharedPreferenceChangeListener;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.Environment;
import de.noisedroid.util.Util;

/**
 * OSM Kacheln werden hier verwaltet, d. h. heruntergeladen/vom Telefonspeicher geladen und zwischengespeichert
 * 
 */
public class TileManager {

	private static TileManager instanz;
	private static LinkedHashMap<String, Bitmap> tileCache = new LinkedHashMap<String, Bitmap>();
	private static ArrayBlockingQueue<String> downloadQueue = new ArrayBlockingQueue<String>(10);
	private static Thread downloadThread;
	private static byte[] downloadBuffer = new byte[2048];
	private static boolean autoDownload;
	private static OnSharedPreferenceChangeListener prefChangeListener = new OnSharedPreferenceChangeListener() {
		public void onSharedPreferenceChanged(SharedPreferences sharedPreferences, String key) {
			if (key.equals("osmAutoDownload")) {
				autoDownload = sharedPreferences.getBoolean("osmAutoDownload", false);
			}
		}
	};

	public static TileManager getInstance() {
		if (instanz == null) {
			instanz = new TileManager();
		}
		return instanz;
	}

	public static void initTileManager(Context context) {
		SharedPreferences mSharedPreferences = context.getSharedPreferences("preferences", Context.MODE_PRIVATE);
		autoDownload = mSharedPreferences.getBoolean("osmAutoDownload", false);
		mSharedPreferences.registerOnSharedPreferenceChangeListener(prefChangeListener);
	}

	public static boolean isAutoDownload() {
		return autoDownload;
	}

	private static void addToCache(String key, Bitmap tile) {
		if (tileCache.size() >= 15) {
			Bitmap remTile = tileCache.remove(tileCache.keySet().iterator().next());
			remTile.recycle();
			remTile = null;
		}
		tileCache.put(key, tile);
	}

	private static void clearCache() {
		for (Entry<String, Bitmap> entry : tileCache.entrySet()) {
			entry.getValue().recycle();
		}
		tileCache.clear();
	}

	private static void addTileDownload(String tileString) {
		if (downloadQueue.contains(tileString)) {
			return;
		}
		if (downloadQueue.remainingCapacity() == 0) {
			downloadQueue.remove();
		}
		downloadQueue.offer(tileString);

		if (downloadThread == null || !downloadThread.isAlive()) {
			downloadThread = new Thread(new Runnable() {
				public void run() {
					while (!downloadQueue.isEmpty()) {
						if (!downloadTile(downloadQueue.poll())) {
							break;
						}
					}
				}
			});
			downloadThread.start();
		}
	}

	public static void clearExternalCache() {
		File osmDir = new File(Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/");
		Util.deleteDirectory(osmDir);
	}

	public static long getCacheSize() {
		File osmDir = new File(Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/");
		return Util.directorySize(osmDir);
	}

	static Bitmap getTile(int x, int y, int z) {
		return getTile(z + "/" + x + "/" + y);
	}

	static boolean tileExists(int x, int y, int z) {
		return tileExists(z + "/" + x + "/" + y);
	}

	static boolean tileExists(String tileString) {
		if (tileCache.containsKey(tileString))
			return true;
		else {
			String filepath = Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/"
					+ tileString + ".png";
			if (new File(filepath).exists())
				return true;
			else
				return false;
		}
	}

	public static void requestTile(String tileString) {
		addTileDownload(tileString);
	}

	public static void requestTile(int x, int y, int z) {
		requestTile(z + "/" + x + "/" + y);
	}

	static Bitmap getTile(String tileString) {
		Bitmap tile = tileCache.get(tileString);
		if (tile != null)
			return tile;

		String filepath = Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/" + tileString
				+ ".png";
		if (new File(filepath).exists()) {
			try {
				tile = BitmapFactory.decodeFile(filepath);
			} catch (OutOfMemoryError e) {
				clearCache();
				return null;
			}
			if (tile != null) {
				addToCache(tileString, tile);
			}
			return tile;
		}
		return null;
	}

	static Bitmap getTile(TileInfo tileInfo) {
		return getTile(tileInfo.x, tileInfo.y, tileInfo.zoom);
	}

	static void downloadTile(int x, int y, int z) {
		downloadTile(z + "/" + x + "/" + y);
	}

	static boolean downloadTile(String tileString) {
		try {
			URL url = new URL("http://tile.openstreetmap.org/" + tileString + ".png");
			URLConnection ucon = url.openConnection();
			InputStream is = ucon.getInputStream();
			// BufferedInputStream bis = new BufferedInputStream(is);
			// ByteArrayBuffer baf = new ByteArrayBuffer(1024);

			int bytesRead = 0;

			File file = new File(Environment.getExternalStorageDirectory() + "/Android/data/de.noisedroid/osm/"
					+ tileString + ".png");
			if (!file.exists()) {
				file.getParentFile().mkdirs();
				file.createNewFile();
			}
			FileOutputStream fos = new FileOutputStream(file);

			while ((bytesRead = is.read(downloadBuffer)) > -1) {
				fos.write(downloadBuffer, 0, bytesRead);
			}

			fos.close();
			is.close();
		} catch (IOException e) {
			return false;
		}
		return true;
	}

	public static boolean tileExists(TileInfo tileInfo) {
		return tileExists(tileInfo.x, tileInfo.y, tileInfo.zoom);
	}
}
