package de.noisedroid.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EventListener;
import java.util.List;

import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.OnSharedPreferenceChangeListener;
import android.location.Criteria;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Bundle;

/**
 * Verwaltet die vom Nutzer angegeben Positionsbestimmungsmethoden und bietet Hilfsmethoden zur Positionsbestimmung
 * 
 */
public class LocationManagerHelper implements OnSharedPreferenceChangeListener {
	/**
	 * ...wobei sich das "better" darauf bezieht, dass er ein neues Positionsevent verkrpert, welches den Anspruch hat
	 * im Sinne der Genauigkeit oder auch zeitlich "besser" zu sein, als ein vorheriges.
	 */
	public interface BetterLocationListener extends EventListener {
		public void onBetterLocation(Location location);
	}

	private Context context;
	private LocationManager mLocationManager;
	private SharedPreferences mSharedPreferences;
	private Location currentBestLocation;
	private boolean aktiv;
	private int minTime, minDistance;
	private List<BetterLocationListener> listenerList = new ArrayList<BetterLocationListener>();
	private LocationListener listener = new LocationListener() {

		public void onStatusChanged(String provider, int status, Bundle extras) {

		}

		public void onProviderEnabled(String provider) {

		}

		public void onProviderDisabled(String provider) {

		}

		public void onLocationChanged(Location location) {
			if (isBetterLocation(location, currentBestLocation)) {
				currentBestLocation = location;
				onBetterLocation(currentBestLocation);
			}
		}
	};

	public LocationManagerHelper(Context context) {
		this.context = context;
		mLocationManager = (LocationManager) this.context.getSystemService(Context.LOCATION_SERVICE);
		mSharedPreferences = context.getSharedPreferences("preferences", Context.MODE_PRIVATE);
		mSharedPreferences.registerOnSharedPreferenceChangeListener(this);
	}

	public String getBestProvider(Criteria criteria, boolean enabledOnly) {
		return mLocationManager.getBestProvider(criteria, enabledOnly);
	}

	public void findBestLocation(int minTime, int minDistance) {
		this.minTime = minTime;
		this.minDistance = minDistance;
		mLocationManager.removeUpdates(listener);
		for (String provider : getProviders()) {
			mLocationManager.requestLocationUpdates(provider, minTime, minDistance, listener);
			Location lastKnown = mLocationManager.getLastKnownLocation(provider);
			if (isBetterLocation(lastKnown, currentBestLocation)) {
				currentBestLocation = lastKnown;
				onBetterLocation(currentBestLocation);
			}
		}
		aktiv = true;
	}

	public Location getBestLastKnownLocation() {
		Location currentBestLastKnown = null;
		for (String provider : getProviders()) {
			Location lastKnown = mLocationManager.getLastKnownLocation(provider);
			if (isBetterLocation(lastKnown, currentBestLastKnown) || currentBestLastKnown == null) {
				currentBestLastKnown = lastKnown;
			}
		}
		return currentBestLastKnown;
	}

	private List<String> getProviders() {
		if (mSharedPreferences.getBoolean("autoProviderAktiv", true)) {
			return mLocationManager.getProviders(false);
		} else {
			String providers = mSharedPreferences.getString("locationProviders", "");
			if (providers.length() == 0) {
				return new ArrayList<String>();
			} else {
				String[] providerArray = providers.split(";");
				return Arrays.asList(providerArray);
			}
		}
	}

	public void stop() {
		mLocationManager.removeUpdates(listener);
		aktiv = false;
	}

	/**
	 * Angelehnt an http://developer.android.com/guide/topics/location/obtaining-user-location.html
	 * 
	 * @param location
	 * @param refLocation
	 * @return
	 */
	private boolean isBetterLocation(Location location, Location refLocation) {
		if (location == null) {
			return false;
		} else if (refLocation == null) {
			// A new location is always better than no location...wenn sie nicht zu alt ist
			long t = location.getTime();
			long t2 = System.currentTimeMillis();
			long timeNowDelta = t2 - t;
			return (timeNowDelta < 30000);
		}
		// Check whether the new location fix is newer or older
		long timeDelta = location.getTime() - refLocation.getTime();
		boolean isSignificantlyNewer = timeDelta > 30000;
		boolean isSignificantlyOlder = timeDelta < -30000;
		boolean isNewer = timeDelta > 0;

		if (isSignificantlyNewer) {
			return true;

		} else if (isSignificantlyOlder) {
			return false;
		}

		// Check whether the new location fix is more or less accurate
		int accuracyDelta = (int) (location.getAccuracy() - refLocation.getAccuracy());
		boolean isLessAccurate = accuracyDelta > 0;
		boolean isMoreAccurate = accuracyDelta < 0;
		boolean isSignificantlyLessAccurate = accuracyDelta > 200;

		// Check if the old and new location are from the same provider
		boolean isFromSameProvider = isSameProvider(location.getProvider(), refLocation.getProvider());

		// Determine location quality using a combination of timeliness and accuracy
		if (isMoreAccurate) {
			return true;
		} else if (isNewer && !isLessAccurate) {
			return true;
		} else if (isNewer && !isSignificantlyLessAccurate && isFromSameProvider) {
			return true;
		}
		return false;
	}

	/**
	 * Von http://developer.android.com/guide/topics/location/obtaining-user-location.html
	 */
	private boolean isSameProvider(String provider1, String provider2) {
		if (provider1 == null) {
			return provider2 == null;
		}
		return provider1.equals(provider2);
	}

	private void onBetterLocation(Location location) {
		for (BetterLocationListener listener : listenerList) {
			listener.onBetterLocation(location);
		}
	}

	public void addBetterLocationListener(BetterLocationListener listener) {
		if (!listenerList.contains(listener)) {
			listenerList.add(listener);
		}
	}

	public void removeBetterLocationListener(BetterLocationListener listener) {
		listenerList.remove(listener);
	}

	public boolean isAktiv() {
		return aktiv;
	}

	public void onSharedPreferenceChanged(SharedPreferences sharedPreferences, String key) {
		if (key.equals("locationProviders") || key.equals("autoProviderAktiv")) {
			if (aktiv) {
				findBestLocation(minTime, minDistance);
			}
		}

	}
}
