package de.noisedroid.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import android.content.Context;
import android.graphics.drawable.Drawable;
import android.graphics.drawable.ShapeDrawable;
import android.graphics.drawable.shapes.OvalShape;
import android.view.Gravity;
import android.view.LayoutInflater;
import android.view.View;
import android.widget.TextView;
import android.widget.Toast;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.ItemizedOverlay;
import com.google.android.maps.OverlayItem;

import de.noisedroid.R;
import de.noisedroid.measure.MeasureAbstract;
import de.noisedroid.measure.MeasureAutomatic;
import de.noisedroid.measure.MeasureSeries;
import de.noisedroid.osm.StaticMapView;

/**
 * Kartenoverlay zur Anzeige von Messungen
 * 
 */
public class MeasureItemOverlay extends ItemizedOverlay<OverlayItem> {

	/**
	 * Einzelnes OverlayItem basierend auf einer Messung
	 * 
	 */
	public class MeasureOverlayItem extends OverlayItem {

		private MeasureAutomatic mMeasure;

		public MeasureOverlayItem(GeoPoint point, MeasureAutomatic measure) {
			this(point, null, null, measure);
		}

		public MeasureOverlayItem(GeoPoint point, String title, String snippet, MeasureAutomatic measure) {
			super(point, title, snippet);
			mMeasure = measure;
			ShapeDrawable drawable = new ShapeDrawable(new OvalShape());
			drawable.getPaint().setColor(Noise.getColorFromNoise(mMeasure.getNoise()));
			drawable.setBounds(-5, -5, 10, 10);

			setMarker(drawable);
		}

		public MeasureAutomatic getMeasure() {
			return mMeasure;
		}

	}

	private ArrayList<MeasureOverlayItem> mOverlays = new ArrayList<MeasureOverlayItem>();
	private Context mContext;
	private Toast infoToast;

	public MeasureItemOverlay(Drawable defaultMarker) {
		super(boundCenter(defaultMarker));
	}

	public MeasureItemOverlay(Context context) {
		super(null);
		mContext = context;
		infoToast = new Toast(mContext);
		populate();
	}

	public MeasureItemOverlay(Collection<MeasureAbstract> measureList, Context context) {
		this(context);
		setMeasures(measureList);
	}

	public void setMeasures(Collection<MeasureAbstract> measureList, MeasureFilter filter) {
		synchronized (mOverlays) {
			mOverlays.clear();
			if (measureList != null) {
				for (MeasureAbstract measure : measureList) {
					if (filter == null || filter.filter(measure, false)) {
						if (measure instanceof MeasureAutomatic) {
							internalAddMeasure((MeasureAutomatic) measure);
						} else if (measure instanceof MeasureSeries) {
							MeasureSeries series = (MeasureSeries) measure;
							for (MeasureAutomatic m : series.getMeasures()) {
								internalAddMeasure(m);
							}
						}
					}
				}
			}
			populate();
		}
	}

	public void setMeasures(Collection<MeasureAbstract> measureList) {
		setMeasures(measureList, null);
	}

	public void addMeasure(MeasureAutomatic measure) {
		synchronized (mOverlays) {
			internalAddMeasure(measure);
			populate();
		}
	}

	private void internalAddMeasure(MeasureAutomatic measure) {
		GeoPoint gp = new GeoPoint((int) (measure.getLocation().getLatitude() * 1E6), (int) (measure.getLocation()
				.getLongitude() * 1E6));
		MeasureOverlayItem overlayitem = new MeasureOverlayItem(gp, measure);
		addOverlay(overlayitem);
	}

	public void removeMeasure(MeasureAutomatic measure) {
		synchronized (mOverlays) {
			Iterator<MeasureOverlayItem> iter = mOverlays.iterator();
			while (iter.hasNext()) {
				if (iter.next().getMeasure() == measure) {
					iter.remove();
				}
			}
			populate();
		}
	}

	public void addOverlay(MeasureOverlayItem overlay) {
		mOverlays.add(overlay);
	}

	@Override
	protected MeasureOverlayItem createItem(int i) {
		synchronized (mOverlays) {
			return mOverlays.get(i);
		}
	}

	@Override
	public int size() {
		synchronized (mOverlays) {
			return mOverlays.size();
		}
	}

	@Override
	protected boolean onTap(int index) {
		synchronized (mOverlays) {
			if (mOverlays.size() != 0) {
				final MeasureOverlayItem item = mOverlays.get(index);

				LayoutInflater inflater = (LayoutInflater) mContext.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
				View layout = inflater.inflate(R.layout.measure_toast, null);
				TextView text = (TextView) layout.findViewById(R.id.text);
				text.setText(item.getMeasure().getNoise() + " dB");
				StaticMapView staticMapView = (StaticMapView) layout.findViewById(R.id.staticmap);

				staticMapView.setPunkt(item.getMeasure().getLocation());

				infoToast.setGravity(Gravity.CENTER_VERTICAL, 0, 0);
				infoToast.setDuration(Toast.LENGTH_LONG);
				infoToast.setView(layout);
				infoToast.show();
				return true;
			} else
				return false;
		}
	}

}
