package de.noiseserver.servlet.app.requestprocessors;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.HashMap;

import javax.mail.MessagingException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sun.org.apache.xerces.internal.dom.DocumentImpl;
import com.sun.org.apache.xml.internal.serialize.XMLSerializer;

import de.noiseserver.Commons;
import de.noiseserver.dto.UserDto;
import de.noiseserver.service.MeasureBeanRemote;
import de.noiseserver.service.MeasureBeanRemote.NotFoundException;
import de.noiseserver.servlet.AuthException;
import de.noiseserver.servlet.NoiseServer;
import de.noiseserver.servlet.app.xmltemplates.XMLSelfdescriptionTemplate;

public class RequestProcessor {
	protected MeasureBeanRemote measureBean;
	private HttpServletRequest req;
	private UserDto user = null;
	private XMLSelfdescriptionTemplate selfdescription;

	// Jeder mgliche Requesttyp ist sozusagen ein eigenes Modul, welches hier
	// "registriert" werden muss um verwendet werden zu knnen
	private final static RequestProcessingElement[] processors = {
			new AddMeasuresProcessor(), new LoginProcessor(),
			new GetMeasuresProcessor(), new UpdateSeriesProcessor(),
			new RegisterProcessor(), new LogoutProcessor(),
			new HasMeasuresProcessor(), new PasswordResetProcessor(), new DeleteUserProcessor() };
	private static HashMap<String, RequestProcessingElement> requestProcessorsMap = new HashMap<String, RequestProcessingElement>();

	static {
		// Initialisierung der einzelnen Request-Processor-Teile
		// Zuordnung des Request-Typs zum Processor
		for (RequestProcessingElement processor : processors) {
			System.out.println("Initializing NoiseDroid Request Processors: '"
					+ processor.getRequestType().toLowerCase() + "'");
			requestProcessorsMap.put(processor.getRequestType().toLowerCase(),
					processor);
		}
	}

	public void processRequest(MeasureBeanRemote measureBean,
			HttpServletRequest req, HttpServletResponse resp)
			throws XMLStreamException, IOException {

		this.req = req;
		this.measureBean = measureBean;

		// Input
		BufferedReader input = req.getReader();
		XMLInputFactory f = XMLInputFactory.newInstance();
		XMLStreamReader reader = f.createXMLStreamReader(input);

		// Output
		Document responseXml = new DocumentImpl();
		Element responseRoot = responseXml
				.createElement(Commons.NODE_RESPONSES_WRAPPER);

		while (reader.hasNext()) {
			int event = reader.next();
			if (event == XMLStreamReader.START_ELEMENT) {
				if (reader.getLocalName()
						.equalsIgnoreCase(Commons.NODE_REQUEST)) {
					// Request-Knoten einem RequestProcessingElement zuordnen
					String type = reader.getAttributeValue(null,
							Commons.ATTR_REQUEST_TYPE).toLowerCase();
					if (requestProcessorsMap.containsKey(type)) {
						requestProcessorsMap.get(type).processRequest(reader,
								responseRoot, measureBean, this);
					} else {
						new UnknownProcessor(type).processRequest(reader,
								responseRoot, measureBean, this);
					}
				} else if (reader.getLocalName().equalsIgnoreCase(
						Commons.NODE_SELFDESCRIPTION)) {
					try {
						JAXBContext context = JAXBContext
								.newInstance(XMLSelfdescriptionTemplate.class);
						Unmarshaller unmarshaller = context
								.createUnmarshaller();
						selfdescription = (XMLSelfdescriptionTemplate) unmarshaller
								.unmarshal(reader);
					} catch (JAXBException e) {
						selfdescription = null;
					}
				}
			} else if (event == XMLStreamReader.END_ELEMENT
					&& reader.getLocalName().equalsIgnoreCase(
							Commons.NODE_REQUESTS_WRAPPER)) {
				// Nicht ber den Requests-Endknoten hinaus bearbeiten
				break;
			}
		}
		responseXml.appendChild(responseRoot);

		XMLSerializer serializer = new XMLSerializer(resp.getWriter(), null);
		serializer.asDOMSerializer();
		serializer.serialize(responseXml.getDocumentElement());
	}

	public void login(String username, String passwordMD5)
			throws AuthException {
		NoiseServer.loginMD5(username, passwordMD5, req);
	}

	public void register(String username, String email, String passwordMD5) throws AuthException {
		NoiseServer.registerMD5(username, email, passwordMD5, req);
	}

	public void logout() {
		NoiseServer.logout(req);
		user = null;
	}

	public boolean isLoggedIn() {
		return req.getUserPrincipal() != null;
	}

	/**
	 * Gibt aktuell angemeldeten, ber Session-Attribut bekannten Nutzer zurck
	 * 
	 * @return
	 */
	public UserDto getUser() {
		if (user == null) {
			Integer userid = (Integer) req.getSession().getAttribute("userid");
			if (userid != null) {
				try {
					user = measureBean.getUserById(userid);
				} catch (NotFoundException e) {
					req.getSession().removeAttribute("userid");
				}
			}
		}
		return user;
	}

	public boolean isUserInRole(String role) {
		return req.isUserInRole(role);
	}

	/**
	 * Gibt an, ob der aktuelle Nutzer in einer mindestens der der angegebenen
	 * Rollen ist.
	 * 
	 * @param roles
	 * @return
	 */
	public boolean isUserInRole(String[] roles) {
		for (String role : roles) {
			if (req.isUserInRole(role)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Gibt die Gertebezeichnung zurck, mit welcher sich das das anfragende
	 * NoiseDroid-Gert auswies.
	 * 
	 * @return
	 */
	public String getDevice() {
		if (selfdescription != null) {
			return selfdescription.device;
		} else {
			return null;
		}
	}

	public void resetPassword(String username, String email) throws AuthException, MessagingException {
		NoiseServer.resetPasswordForUser(username, email, req);
	}
	public void deleteUser(String username, String password) throws AuthException{
		NoiseServer.deleteUser(username, password, req);
	}
}
